"""Tests for SequenceVar class."""

import pytest

from pycsp3_scheduling.variables import (
    IntervalVar,
    IntervalVarArray,
    SequenceVar,
    SequenceVarArray,
    clear_interval_registry,
    clear_sequence_registry,
)


@pytest.fixture(autouse=True)
def reset_registries():
    """Reset registries before each test."""
    clear_interval_registry()
    clear_sequence_registry()
    yield
    clear_interval_registry()
    clear_sequence_registry()


class TestSequenceVar:
    """Tests for SequenceVar class."""

    def test_basic_creation(self):
        """Test basic sequence variable creation."""
        task1 = IntervalVar(size=10, name="t1")
        task2 = IntervalVar(size=15, name="t2")
        task3 = IntervalVar(size=8, name="t3")
        seq = SequenceVar(intervals=[task1, task2, task3], name="machine1")

        assert seq.name == "machine1"
        assert seq.size == 3
        assert len(seq) == 3
        assert not seq.has_types

    def test_with_types(self):
        """Test sequence with type identifiers."""
        task1 = IntervalVar(size=10, name="t1")
        task2 = IntervalVar(size=15, name="t2")
        task3 = IntervalVar(size=8, name="t3")
        seq = SequenceVar(
            intervals=[task1, task2, task3],
            types=[0, 1, 0],
            name="machine1"
        )

        assert seq.has_types
        assert seq.get_type(0) == 0
        assert seq.get_type(1) == 1
        assert seq.get_type(2) == 0

    def test_get_intervals_by_type(self):
        """Test filtering intervals by type."""
        task1 = IntervalVar(size=10, name="t1")
        task2 = IntervalVar(size=15, name="t2")
        task3 = IntervalVar(size=8, name="t3")
        seq = SequenceVar(
            intervals=[task1, task2, task3],
            types=[0, 1, 0],
            name="machine1"
        )

        type0_intervals = seq.get_intervals_by_type(0)
        assert len(type0_intervals) == 2
        assert task1 in type0_intervals
        assert task3 in type0_intervals

        type1_intervals = seq.get_intervals_by_type(1)
        assert len(type1_intervals) == 1
        assert task2 in type1_intervals

    def test_indexing(self):
        """Test interval access by index."""
        task1 = IntervalVar(size=10, name="t1")
        task2 = IntervalVar(size=15, name="t2")
        seq = SequenceVar(intervals=[task1, task2], name="seq")

        assert seq[0] == task1
        assert seq[1] == task2
        assert seq.get_interval(0) == task1

    def test_iteration(self):
        """Test iteration over intervals."""
        tasks = [IntervalVar(size=10, name=f"t{i}") for i in range(3)]
        seq = SequenceVar(intervals=tasks, name="seq")

        for i, interval in enumerate(seq):
            assert interval == tasks[i]

    def test_auto_name_generation(self):
        """Test automatic name generation."""
        task = IntervalVar(size=10, name="t")
        seq1 = SequenceVar(intervals=[task])
        seq2 = SequenceVar(intervals=[task])

        assert seq1.name != seq2.name
        assert seq1.name.startswith("_sequence_")

    def test_unique_ids(self):
        """Test that each sequence gets a unique ID."""
        task = IntervalVar(size=10, name="t")
        seq1 = SequenceVar(intervals=[task], name="s1")
        seq2 = SequenceVar(intervals=[task], name="s2")

        assert seq1._id != seq2._id

    def test_equality_by_id(self):
        """Test that equality is based on ID."""
        task = IntervalVar(size=10, name="t")
        seq1 = SequenceVar(intervals=[task], name="seq")
        seq2 = SequenceVar(intervals=[task], name="seq")  # Same name, different ID

        assert seq1 != seq2
        assert seq1 == seq1

    def test_hashable(self):
        """Test that sequences are hashable."""
        task = IntervalVar(size=10, name="t")
        seq1 = SequenceVar(intervals=[task], name="s1")
        seq2 = SequenceVar(intervals=[task], name="s2")

        s = {seq1, seq2}
        assert len(s) == 2

    def test_empty_sequence(self):
        """Test empty sequence creation."""
        seq = SequenceVar(intervals=[], name="empty")
        assert seq.size == 0
        assert len(seq) == 0

    def test_types_length_mismatch(self):
        """Test error when types length doesn't match intervals."""
        task1 = IntervalVar(size=10, name="t1")
        task2 = IntervalVar(size=15, name="t2")

        with pytest.raises(ValueError, match="Length of types"):
            SequenceVar(
                intervals=[task1, task2],
                types=[0],  # Wrong length
                name="seq"
            )

    def test_negative_type(self):
        """Test error on negative type identifier."""
        task = IntervalVar(size=10, name="t")

        with pytest.raises(ValueError, match="non-negative integer"):
            SequenceVar(
                intervals=[task],
                types=[-1],
                name="seq"
            )

    def test_repr(self):
        """Test string representation."""
        task1 = IntervalVar(size=10, name="t1")
        task2 = IntervalVar(size=15, name="t2")
        seq = SequenceVar(intervals=[task1, task2], name="machine1")

        repr_str = repr(seq)
        assert "SequenceVar" in repr_str
        assert "machine1" in repr_str
        assert "t1" in repr_str
        assert "t2" in repr_str


class TestSequenceVarArray:
    """Tests for SequenceVarArray factory function."""

    def test_array_with_intervals(self):
        """Test array creation with intervals per sequence."""
        ops_m0 = IntervalVarArray(3, size_range=10, name="op_m0")
        ops_m1 = IntervalVarArray(3, size_range=15, name="op_m1")

        sequences = SequenceVarArray(
            2,
            intervals_per_sequence=[ops_m0, ops_m1],
            name="machine"
        )

        assert len(sequences) == 2
        assert sequences[0].name == "machine[0]"
        assert sequences[1].name == "machine[1]"
        assert sequences[0].size == 3
        assert sequences[1].size == 3

    def test_array_with_types(self):
        """Test array creation with types per sequence."""
        ops_m0 = IntervalVarArray(3, size_range=10, name="op_m0")
        ops_m1 = IntervalVarArray(3, size_range=15, name="op_m1")

        sequences = SequenceVarArray(
            2,
            intervals_per_sequence=[ops_m0, ops_m1],
            types_per_sequence=[[0, 1, 0], [1, 0, 1]],
            name="machine"
        )

        assert sequences[0].has_types
        assert sequences[1].has_types
        assert sequences[0].get_type(0) == 0
        assert sequences[1].get_type(0) == 1

    def test_empty_array(self):
        """Test array without intervals."""
        sequences = SequenceVarArray(3, name="machine")
        assert len(sequences) == 3
        for seq in sequences:
            assert seq.size == 0

    def test_size_mismatch_error(self):
        """Test error when intervals_per_sequence size doesn't match."""
        ops = IntervalVarArray(3, size_range=10, name="op")

        with pytest.raises(ValueError, match="must match size"):
            SequenceVarArray(
                2,
                intervals_per_sequence=[ops],  # Only 1, but size is 2
                name="machine"
            )



"""
pycsp3-scheduling: Scheduling extension for pycsp3.

This package extends pycsp3 with comprehensive scheduling support including:
- Interval variables (IntervalVar)
- Sequence variables (SequenceVar)
- Precedence constraints (end_before_start, start_at_start, etc.)
- Grouping constraints (span, alternative, synchronize)
- Sequence constraints (SeqNoOverlap, first, last, before, previous)
- Sequence consistency constraints (same_sequence, same_common_subsequence)
- Sequence accessor expressions (start_of_next, start_of_prev, etc.)
- Cumulative functions (pulse, step_at_start, step_at_end, Cumulative)
- State functions (StateFunction, TransitionMatrix, always_equal, etc.)
- XCSP3 scheduling extension output [future]

"""

from __future__ import annotations

__version__ = "0.1.0"
__author__ = "Sohaib AFIFI"

# Variables
from pycsp3_scheduling.variables import (
    INTERVAL_MAX,
    INTERVAL_MIN,
    IntervalVar,
    IntervalVarArray,
    IntervalVarDict,
    SequenceVar,
    SequenceVarArray,
)

# Expressions
from pycsp3_scheduling.expressions import (
    IntervalExpr,
    end_of,
    end_of_next,
    end_of_prev,
    expr_max,
    expr_min,
    length_of,
    length_of_next,
    length_of_prev,
    overlap_length,
    presence_of,
    size_of,
    size_of_next,
    size_of_prev,
    start_of,
    start_of_next,
    start_of_prev,
    type_of_next,
    type_of_prev,
)

# Interop
from pycsp3_scheduling.interop import end_time, interval_value, presence_time, start_time

# Constraints
from pycsp3_scheduling.constraints import (
    SeqCumulative,
    SeqNoOverlap,
    alternative,
    before,
    end_at_end,
    end_at_start,
    end_before_end,
    end_before_start,
    first,
    last,
    previous,
    same_common_subsequence,
    same_sequence,
    span,
    start_at_end,
    start_at_start,
    start_before_end,
    start_before_start,
    synchronize,
)

# Cumulative Functions
from pycsp3_scheduling.functions import (
    CumulExpr,
    CumulFunction,
    CumulConstraint,
    CumulHeightExpr,
    always_in,
    cumul_range,
    height_at_end,
    height_at_start,
    pulse,
    step_at,
    step_at_end,
    step_at_start,
)

# State Functions
from pycsp3_scheduling.functions import (
    StateFunction,
    StateConstraint,
    TransitionMatrix,
    always_constant,
    always_equal,
    always_no_state,
)

__all__ = [
    "__version__",
    # Variables
    "IntervalVar",
    "IntervalVarArray",
    "IntervalVarDict",
    "SequenceVar",
    "SequenceVarArray",
    "INTERVAL_MIN",
    "INTERVAL_MAX",
    # Expressions - Basic
    "IntervalExpr",
    "start_of",
    "end_of",
    "size_of",
    "length_of",
    "presence_of",
    "overlap_length",
    "expr_min",
    "expr_max",
    # Expressions - Sequence Next
    "start_of_next",
    "end_of_next",
    "size_of_next",
    "length_of_next",
    "type_of_next",
    # Expressions - Sequence Prev
    "start_of_prev",
    "end_of_prev",
    "size_of_prev",
    "length_of_prev",
    "type_of_prev",
    # Interop
    "start_time",
    "end_time",
    "presence_time",
    "interval_value",
    # Constraints - Exact timing
    "start_at_start",
    "start_at_end",
    "end_at_start",
    "end_at_end",
    # Constraints - Before
    "start_before_start",
    "start_before_end",
    "end_before_start",
    "end_before_end",
    # Constraints - Grouping
    "span",
    "alternative",
    "synchronize",
    # Constraints - Sequence
    "SeqNoOverlap",
    "first",
    "last",
    "before",
    "previous",
    # Constraints - Sequence Consistency
    "same_sequence",
    "same_common_subsequence",
    # Constraints - Cumulative
    "SeqCumulative",
    # Cumulative Functions
    "CumulExpr",
    "CumulFunction",
    "CumulConstraint",
    "CumulHeightExpr",
    "pulse",
    "step_at",
    "step_at_start",
    "step_at_end",
    "cumul_range",
    "always_in",
    "height_at_start",
    "height_at_end",
    # State Functions
    "StateFunction",
    "StateConstraint",
    "TransitionMatrix",
    "always_equal",
    "always_constant",
    "always_no_state",
]



"""
Aircraft Landing Problem

This model uses pycsp3-scheduling's IntervalVar and pycsp3's NoOverlap constraint
to schedule aircraft landings while respecting separation requirements.

## Data Example
  airland01.json

## Model
  variables: IntervalVar
  constraints: NoOverlap (pycsp3)

## Execution
  python AircraftLanding.py -data=<datafile.json>

## Links
  - http://people.brunel.ac.uk/~mastjjb/jeb/orlib/airlandinfo.html
  - https://www.jstor.org/stable/25768908
  - https://github.com/xcsp3team/PyCSP3-models/blob/main/realistic/AircraftLanding/AircraftLanding.py

## Tags
  realistic, scheduling, xcsp22
"""

import json
import os
from itertools import combinations

from pycsp3 import (
    Var,
    VarArray,
    AllDifferent,
    NoOverlap,
    satisfy,
    minimize,
    solve,
    SAT,
    OPTIMUM,
    value,
)
from pycsp3_scheduling import (
    IntervalVar,
    start_time,
    interval_value,
)

# Load data
_data_dir = os.path.join(os.path.dirname(__file__), "data")
with open(os.path.join(_data_dir, "airland01.json")) as f:
    _data = json.load(f)

nPlanes = _data["nPlanes"]
times = [(t["earliest"], t["target"], t["latest"]) for t in _data["times"]]
costs = [(c["early_penalty"], c["late_penalty"]) for c in _data["costs"]]
separations = _data["separations"]

earliest, target, latest = zip(*times)
early_penalties, late_penalties = zip(*costs)

P = range(nPlanes)

# landing[i] is the interval for plane i's landing (size=1 represents the landing event)
landing = [
    IntervalVar(
        start=(earliest[i], latest[i]),
        size=1,  # Landing is a point event with size 1
        name=f"plane_{i}",
    )
    for i in P
]

# x[i] is the landing time of plane i (for easier access)
x = VarArray(size=nPlanes, dom=lambda i: range(earliest[i], latest[i] + 1))

# erl[i] is the earliness of plane i
erl = VarArray(size=nPlanes, dom=lambda i: range(target[i] - earliest[i] + 1))

# trd[i] is the tardiness of plane i
trd = VarArray(size=nPlanes, dom=lambda i: range(latest[i] - target[i] + 1))

satisfy(
    # link landing interval to x variable
    [start_time(landing[i]) == x[i] for i in P],
    # planes must land at different times
    AllDifferent(x),
    # separation constraints using NoOverlap
    [
        NoOverlap(
            origins=[x[i], x[j]],
            lengths=[separations[i][j], separations[j][i]],
        )
        for i, j in combinations(P, 2)
    ],
    # computing earliness of planes
    [erl[i] == max(0, target[i] - x[i]) for i in P],
    # computing tardiness of planes
    [trd[i] == max(0, x[i] - target[i]) for i in P],
)

minimize(
    # minimizing the deviation cost
    erl * early_penalties + trd * late_penalties
)

# --- Solution output ---
if __name__ == "__main__":
    if solve() in (SAT, OPTIMUM):
        total_cost = sum(
            value(erl[i]) * early_penalties[i] + value(trd[i]) * late_penalties[i]
            for i in P
        )
        print(f"Total deviation cost: {total_cost}")
        print("\nLanding schedule:")
        for i in P:
            v = interval_value(landing[i])
            early = value(erl[i])
            late = value(trd[i])
            status = f"early={early}" if early > 0 else (f"late={late}" if late > 0 else "on-time")
            print(
                f"  Plane {i}: land at {v['start']} (target={target[i]}, window=[{earliest[i]},{latest[i]}]) {status}"
            )



"""Tests for sequence constraints and expressions (Phase 5)."""

import pytest

pycsp3 = pytest.importorskip("pycsp3")

from pycsp3.classes.entities import ECtr, clear as clear_pycsp3
from pycsp3.classes.main.constraints import ConstraintNoOverlap
from pycsp3.classes.nodes import Node, TypeNode

from pycsp3_scheduling.constraints import (
    SeqNoOverlap,
    before,
    first,
    last,
    previous,
    same_common_subsequence,
    same_sequence,
)
from pycsp3_scheduling.constraints._pycsp3 import clear_pycsp3_cache
from pycsp3_scheduling.expressions import (
    end_of_next,
    end_of_prev,
    length_of_next,
    length_of_prev,
    size_of_next,
    size_of_prev,
    start_of_next,
    start_of_prev,
    type_of_next,
    type_of_prev,
)
from pycsp3_scheduling.variables import (
    IntervalVar,
    SequenceVar,
    clear_interval_registry,
    clear_sequence_registry,
)


@pytest.fixture(autouse=True)
def reset_state():
    """Reset registries and caches before each test."""
    clear_pycsp3()
    clear_interval_registry()
    clear_sequence_registry()
    clear_pycsp3_cache()
    yield
    clear_pycsp3()
    clear_interval_registry()
    clear_sequence_registry()
    clear_pycsp3_cache()


# =============================================================================
# SeqNoOverlap Tests
# =============================================================================


class TestSeqNoOverlap:
    """Tests for SeqNoOverlap constraint with transition matrix."""

    def test_basic_no_overlap(self):
        """Basic SeqNoOverlap without transitions."""
        tasks = [IntervalVar(size=5, name=f"t{i}") for i in range(3)]
        ctr = SeqNoOverlap(tasks)
        assert isinstance(ctr, ECtr)
        assert isinstance(ctr.constraint, ConstraintNoOverlap)

    def test_empty_sequence(self):
        """Empty sequence returns empty constraint list."""
        result = SeqNoOverlap([])
        assert result == []

    def test_with_sequence_var(self):
        """SeqNoOverlap accepts SequenceVar."""
        tasks = [IntervalVar(size=5, name=f"t{i}") for i in range(3)]
        seq = SequenceVar(intervals=tasks, name="machine")
        ctr = SeqNoOverlap(seq)
        assert isinstance(ctr, ECtr)

    def test_transition_matrix_basic(self):
        """SeqNoOverlap with transition matrix returns list of constraints."""
        tasks = [IntervalVar(size=5, name=f"t{i}") for i in range(3)]
        seq = SequenceVar(intervals=tasks, types=[0, 1, 0], name="machine")
        matrix = [[0, 5], [3, 0]]  # Setup times between types
        
        result = SeqNoOverlap(seq, transition_matrix=matrix)
        
        assert isinstance(result, list)
        assert len(result) > 1  # NoOverlap + transition constraints

    def test_transition_matrix_requires_sequence_var(self):
        """transition_matrix requires SequenceVar with types."""
        tasks = [IntervalVar(size=5, name=f"t{i}") for i in range(3)]
        matrix = [[0, 5], [3, 0]]
        
        with pytest.raises(ValueError, match="requires a SequenceVar with types"):
            SeqNoOverlap(tasks, transition_matrix=matrix)

    def test_transition_matrix_requires_types(self):
        """transition_matrix requires SequenceVar with types defined."""
        tasks = [IntervalVar(size=5, name=f"t{i}") for i in range(3)]
        seq = SequenceVar(intervals=tasks, name="machine")  # No types
        matrix = [[0, 5], [3, 0]]
        
        with pytest.raises(ValueError, match="requires a SequenceVar with types"):
            SeqNoOverlap(seq, transition_matrix=matrix)

    def test_transition_matrix_dimension_validation(self):
        """transition_matrix dimensions must match types."""
        tasks = [IntervalVar(size=5, name=f"t{i}") for i in range(3)]
        seq = SequenceVar(intervals=tasks, types=[0, 1, 2], name="machine")
        matrix = [[0, 5], [3, 0]]  # Only 2 types, but we have type 2
        
        with pytest.raises(ValueError, match="at least 3 rows"):
            SeqNoOverlap(seq, transition_matrix=matrix)

    def test_transition_matrix_row_validation(self):
        """transition_matrix rows must be lists."""
        tasks = [IntervalVar(size=5, name=f"t{i}") for i in range(2)]
        seq = SequenceVar(intervals=tasks, types=[0, 1], name="machine")
        matrix = [[0, 5], "invalid"]
        
        with pytest.raises(TypeError, match="must be a list"):
            SeqNoOverlap(seq, transition_matrix=matrix)

    def test_transition_matrix_column_validation(self):
        """transition_matrix columns must have correct length."""
        tasks = [IntervalVar(size=5, name=f"t{i}") for i in range(2)]
        seq = SequenceVar(intervals=tasks, types=[0, 1], name="machine")
        matrix = [[0], [3, 0]]  # First row too short
        
        with pytest.raises(ValueError, match="at least 2 columns"):
            SeqNoOverlap(seq, transition_matrix=matrix)

    def test_transition_matrix_type_validation(self):
        """transition_matrix must be a list."""
        tasks = [IntervalVar(size=5, name=f"t{i}") for i in range(2)]
        seq = SequenceVar(intervals=tasks, types=[0, 1], name="machine")
        
        with pytest.raises(TypeError, match="must be a 2D list"):
            SeqNoOverlap(seq, transition_matrix="not a list")


# =============================================================================
# Sequence Ordering Constraints Tests
# =============================================================================


class TestFirst:
    """Tests for first() constraint."""

    def test_basic_first(self):
        """first() returns list of constraints."""
        tasks = [IntervalVar(size=5, name=f"t{i}") for i in range(3)]
        seq = SequenceVar(intervals=tasks, name="machine")
        
        constraints = first(seq, tasks[0])
        
        assert isinstance(constraints, list)
        assert len(constraints) == 2  # Against t1 and t2
        for c in constraints:
            assert isinstance(c, Node)
            assert c.type == TypeNode.LE

    def test_first_with_list(self):
        """first() works with list of intervals."""
        tasks = [IntervalVar(size=5, name=f"t{i}") for i in range(3)]
        
        constraints = first(tasks, tasks[0])
        
        assert isinstance(constraints, list)
        assert len(constraints) == 2

    def test_first_single_interval(self):
        """first() with single interval returns empty list."""
        task = IntervalVar(size=5, name="t0")
        
        constraints = first([task], task)
        
        assert constraints == []

    def test_first_optional_interval(self):
        """first() with optional intervals includes presence conditions."""
        tasks = [
            IntervalVar(size=5, optional=True, name="t0"),
            IntervalVar(size=5, name="t1"),
        ]
        
        constraints = first(tasks, tasks[0])
        
        assert isinstance(constraints, list)
        for c in constraints:
            assert isinstance(c, Node)
            # Should be OR constraint (absence or ordering)
            assert c.type == TypeNode.OR

    def test_first_interval_not_in_sequence(self):
        """first() raises error for interval not in sequence."""
        tasks = [IntervalVar(size=5, name=f"t{i}") for i in range(2)]
        other = IntervalVar(size=5, name="other")
        
        with pytest.raises(ValueError, match="not in the sequence"):
            first(tasks, other)

    def test_first_invalid_interval_type(self):
        """first() raises error for non-IntervalVar."""
        tasks = [IntervalVar(size=5, name="t0")]
        
        with pytest.raises(TypeError, match="must be an IntervalVar"):
            first(tasks, "not_interval")


class TestLast:
    """Tests for last() constraint."""

    def test_basic_last(self):
        """last() returns list of constraints."""
        tasks = [IntervalVar(size=5, name=f"t{i}") for i in range(3)]
        seq = SequenceVar(intervals=tasks, name="machine")
        
        constraints = last(seq, tasks[2])
        
        assert isinstance(constraints, list)
        assert len(constraints) == 2  # Against t0 and t1
        for c in constraints:
            assert isinstance(c, Node)
            assert c.type == TypeNode.LE

    def test_last_single_interval(self):
        """last() with single interval returns empty list."""
        task = IntervalVar(size=5, name="t0")
        
        constraints = last([task], task)
        
        assert constraints == []

    def test_last_optional_intervals(self):
        """last() with optional intervals includes presence conditions."""
        tasks = [
            IntervalVar(size=5, name="t0"),
            IntervalVar(size=5, optional=True, name="t1"),
        ]
        
        constraints = last(tasks, tasks[1])
        
        assert isinstance(constraints, list)
        for c in constraints:
            assert c.type == TypeNode.OR

    def test_last_interval_not_in_sequence(self):
        """last() raises error for interval not in sequence."""
        tasks = [IntervalVar(size=5, name="t0")]
        other = IntervalVar(size=5, name="other")
        
        with pytest.raises(ValueError, match="not in the sequence"):
            last(tasks, other)


class TestBefore:
    """Tests for before() constraint."""

    def test_basic_before(self):
        """before() returns precedence constraint."""
        tasks = [IntervalVar(size=5, name=f"t{i}") for i in range(3)]
        seq = SequenceVar(intervals=tasks, name="machine")
        
        constraints = before(seq, tasks[0], tasks[2])
        
        assert isinstance(constraints, list)
        assert len(constraints) == 1
        assert isinstance(constraints[0], Node)
        assert constraints[0].type == TypeNode.LE

    def test_before_same_interval(self):
        """before() raises error for same interval."""
        task = IntervalVar(size=5, name="t0")
        
        with pytest.raises(ValueError, match="must be different"):
            before([task, task], task, task)

    def test_before_optional_intervals(self):
        """before() with optional intervals includes presence conditions."""
        tasks = [
            IntervalVar(size=5, optional=True, name="t0"),
            IntervalVar(size=5, optional=True, name="t1"),
        ]
        
        constraints = before(tasks, tasks[0], tasks[1])
        
        assert len(constraints) == 1
        assert constraints[0].type == TypeNode.OR

    def test_before_interval_not_in_sequence(self):
        """before() raises error for interval not in sequence."""
        tasks = [IntervalVar(size=5, name=f"t{i}") for i in range(2)]
        other = IntervalVar(size=5, name="other")
        
        with pytest.raises(ValueError, match="not in the sequence"):
            before(tasks, tasks[0], other)


class TestPrevious:
    """Tests for previous() constraint."""

    def test_basic_previous(self):
        """previous() returns precedence + no-between constraints."""
        tasks = [IntervalVar(size=5, name=f"t{i}") for i in range(3)]
        seq = SequenceVar(intervals=tasks, name="machine")
        
        constraints = previous(seq, tasks[0], tasks[1])
        
        assert isinstance(constraints, list)
        # 1 precedence + 1 no-between (for t2)
        assert len(constraints) == 2

    def test_previous_two_intervals(self):
        """previous() with only two intervals."""
        tasks = [IntervalVar(size=5, name=f"t{i}") for i in range(2)]
        
        constraints = previous(tasks, tasks[0], tasks[1])
        
        # Just precedence, no third interval to be "between"
        assert len(constraints) == 1

    def test_previous_same_interval(self):
        """previous() raises error for same interval."""
        task = IntervalVar(size=5, name="t0")
        
        with pytest.raises(ValueError, match="must be different"):
            previous([task, task], task, task)

    def test_previous_optional_intervals(self):
        """previous() with optional intervals."""
        tasks = [
            IntervalVar(size=5, optional=True, name="t0"),
            IntervalVar(size=5, optional=True, name="t1"),
            IntervalVar(size=5, optional=True, name="t2"),
        ]
        
        constraints = previous(tasks, tasks[0], tasks[1])
        
        assert isinstance(constraints, list)
        for c in constraints:
            assert c.type == TypeNode.OR


# =============================================================================
# Sequence Consistency Constraints Tests
# =============================================================================


class TestSameSequence:
    """Tests for same_sequence() constraint."""

    def test_basic_same_sequence(self):
        """same_sequence() returns ordering constraints."""
        tasks = [IntervalVar(size=5, name=f"t{i}") for i in range(3)]
        seq1 = SequenceVar(intervals=tasks, name="machine1")
        seq2 = SequenceVar(intervals=tasks, name="machine2")
        
        constraints = same_sequence(seq1, seq2)
        
        assert isinstance(constraints, list)
        # 3 common intervals = 3 choose 2 = 3 pairs
        assert len(constraints) == 3

    def test_same_sequence_no_common(self):
        """same_sequence() with no common intervals returns empty."""
        tasks1 = [IntervalVar(size=5, name=f"t1_{i}") for i in range(2)]
        tasks2 = [IntervalVar(size=5, name=f"t2_{i}") for i in range(2)]
        
        constraints = same_sequence(tasks1, tasks2)
        
        assert constraints == []

    def test_same_sequence_one_common(self):
        """same_sequence() with one common interval returns empty."""
        common = IntervalVar(size=5, name="common")
        tasks1 = [common, IntervalVar(size=5, name="t1")]
        tasks2 = [common, IntervalVar(size=5, name="t2")]
        
        constraints = same_sequence(tasks1, tasks2)
        
        assert constraints == []

    def test_same_sequence_partial_overlap(self):
        """same_sequence() with partial overlap."""
        t0 = IntervalVar(size=5, name="t0")
        t1 = IntervalVar(size=5, name="t1")
        t2 = IntervalVar(size=5, name="t2")
        t3 = IntervalVar(size=5, name="t3")
        
        # t0, t1 are common
        tasks1 = [t0, t1, t2]
        tasks2 = [t0, t1, t3]
        
        constraints = same_sequence(tasks1, tasks2)
        
        # 2 common intervals = 1 pair
        assert len(constraints) == 1


class TestSameCommonSubsequence:
    """Tests for same_common_subsequence() constraint."""

    def test_basic_same_common_subsequence(self):
        """same_common_subsequence() returns ordering constraints."""
        tasks = [IntervalVar(size=5, name=f"t{i}") for i in range(3)]
        seq1 = SequenceVar(intervals=tasks, name="machine1")
        seq2 = SequenceVar(intervals=tasks, name="machine2")
        
        constraints = same_common_subsequence(seq1, seq2)
        
        assert isinstance(constraints, list)
        assert len(constraints) == 3  # 3 pairs

    def test_same_common_subsequence_no_common(self):
        """same_common_subsequence() with no common intervals returns empty."""
        tasks1 = [IntervalVar(size=5, name=f"t1_{i}") for i in range(2)]
        tasks2 = [IntervalVar(size=5, name=f"t2_{i}") for i in range(2)]
        
        constraints = same_common_subsequence(tasks1, tasks2)
        
        assert constraints == []

    def test_same_common_subsequence_optional(self):
        """same_common_subsequence() with optional intervals."""
        tasks = [IntervalVar(size=5, optional=True, name=f"t{i}") for i in range(3)]
        
        constraints = same_common_subsequence(tasks, tasks)
        
        assert isinstance(constraints, list)
        for c in constraints:
            assert c.type == TypeNode.OR


# =============================================================================
# Sequence Accessor Expressions Tests
# =============================================================================


class TestSequenceAccessorNext:
    """Tests for sequence accessor expressions (next)."""

    def test_start_of_next_basic(self):
        """start_of_next returns IntervalExpr."""
        from pycsp3_scheduling.expressions import IntervalExpr
        
        tasks = [IntervalVar(size=5, name=f"t{i}") for i in range(3)]
        seq = SequenceVar(intervals=tasks, name="machine")
        
        expr = start_of_next(seq, tasks[0])
        
        assert isinstance(expr, IntervalExpr)

    def test_end_of_next_basic(self):
        """end_of_next returns IntervalExpr."""
        from pycsp3_scheduling.expressions import IntervalExpr
        
        tasks = [IntervalVar(size=5, name=f"t{i}") for i in range(3)]
        
        expr = end_of_next(tasks, tasks[0])
        
        assert isinstance(expr, IntervalExpr)

    def test_size_of_next_basic(self):
        """size_of_next returns IntervalExpr."""
        from pycsp3_scheduling.expressions import IntervalExpr
        
        tasks = [IntervalVar(size=5, name=f"t{i}") for i in range(3)]
        
        expr = size_of_next(tasks, tasks[0])
        
        assert isinstance(expr, IntervalExpr)

    def test_length_of_next_basic(self):
        """length_of_next returns IntervalExpr."""
        from pycsp3_scheduling.expressions import IntervalExpr
        
        tasks = [IntervalVar(size=5, name=f"t{i}") for i in range(3)]
        
        expr = length_of_next(tasks, tasks[0])
        
        assert isinstance(expr, IntervalExpr)

    def test_type_of_next_basic(self):
        """type_of_next returns IntervalExpr."""
        from pycsp3_scheduling.expressions import IntervalExpr
        
        tasks = [IntervalVar(size=5, name=f"t{i}") for i in range(3)]
        seq = SequenceVar(intervals=tasks, types=[0, 1, 2], name="machine")
        
        expr = type_of_next(seq, tasks[0])
        
        assert isinstance(expr, IntervalExpr)

    def test_type_of_next_requires_sequence_var(self):
        """type_of_next requires SequenceVar."""
        tasks = [IntervalVar(size=5, name=f"t{i}") for i in range(3)]
        
        with pytest.raises(TypeError, match="requires a SequenceVar"):
            type_of_next(tasks, tasks[0])

    def test_type_of_next_requires_types(self):
        """type_of_next requires types defined."""
        tasks = [IntervalVar(size=5, name=f"t{i}") for i in range(3)]
        seq = SequenceVar(intervals=tasks, name="machine")
        
        with pytest.raises(ValueError, match="requires sequence with types"):
            type_of_next(seq, tasks[0])

    def test_accessor_interval_not_in_sequence(self):
        """Accessors raise error for interval not in sequence."""
        tasks = [IntervalVar(size=5, name=f"t{i}") for i in range(2)]
        other = IntervalVar(size=5, name="other")
        
        with pytest.raises(ValueError, match="not in the sequence"):
            start_of_next(tasks, other)


class TestSequenceAccessorPrev:
    """Tests for sequence accessor expressions (prev)."""

    def test_start_of_prev_basic(self):
        """start_of_prev returns IntervalExpr."""
        from pycsp3_scheduling.expressions import IntervalExpr
        
        tasks = [IntervalVar(size=5, name=f"t{i}") for i in range(3)]
        
        expr = start_of_prev(tasks, tasks[1])
        
        assert isinstance(expr, IntervalExpr)

    def test_end_of_prev_basic(self):
        """end_of_prev returns IntervalExpr."""
        from pycsp3_scheduling.expressions import IntervalExpr
        
        tasks = [IntervalVar(size=5, name=f"t{i}") for i in range(3)]
        
        expr = end_of_prev(tasks, tasks[1])
        
        assert isinstance(expr, IntervalExpr)

    def test_size_of_prev_basic(self):
        """size_of_prev returns IntervalExpr."""
        from pycsp3_scheduling.expressions import IntervalExpr
        
        tasks = [IntervalVar(size=5, name=f"t{i}") for i in range(3)]
        
        expr = size_of_prev(tasks, tasks[1])
        
        assert isinstance(expr, IntervalExpr)

    def test_length_of_prev_basic(self):
        """length_of_prev returns IntervalExpr."""
        from pycsp3_scheduling.expressions import IntervalExpr
        
        tasks = [IntervalVar(size=5, name=f"t{i}") for i in range(3)]
        
        expr = length_of_prev(tasks, tasks[1])
        
        assert isinstance(expr, IntervalExpr)

    def test_type_of_prev_basic(self):
        """type_of_prev returns IntervalExpr."""
        from pycsp3_scheduling.expressions import IntervalExpr
        
        tasks = [IntervalVar(size=5, name=f"t{i}") for i in range(3)]
        seq = SequenceVar(intervals=tasks, types=[0, 1, 2], name="machine")
        
        expr = type_of_prev(seq, tasks[1])
        
        assert isinstance(expr, IntervalExpr)

    def test_type_of_prev_requires_sequence_var(self):
        """type_of_prev requires SequenceVar."""
        tasks = [IntervalVar(size=5, name=f"t{i}") for i in range(3)]
        
        with pytest.raises(TypeError, match="requires a SequenceVar"):
            type_of_prev(tasks, tasks[1])

    def test_type_of_prev_requires_types(self):
        """type_of_prev requires types defined."""
        tasks = [IntervalVar(size=5, name=f"t{i}") for i in range(3)]
        seq = SequenceVar(intervals=tasks, name="machine")
        
        with pytest.raises(ValueError, match="requires sequence with types"):
            type_of_prev(seq, tasks[1])


# =============================================================================
# Integration Tests
# =============================================================================


class TestSequenceIntegration:
    """Integration tests for sequence constraints with pycsp3."""

    def test_satisfy_with_first_last(self):
        """Test first and last work with satisfy()."""
        from pycsp3 import satisfy
        
        tasks = [IntervalVar(size=5, name=f"t{i}") for i in range(3)]
        seq = SequenceVar(intervals=tasks, name="machine")
        
        satisfy(first(seq, tasks[0]))
        satisfy(last(seq, tasks[2]))

    def test_satisfy_with_before(self):
        """Test before works with satisfy()."""
        from pycsp3 import satisfy
        
        tasks = [IntervalVar(size=5, name=f"t{i}") for i in range(3)]
        
        satisfy(before(tasks, tasks[0], tasks[2]))

    def test_satisfy_with_previous(self):
        """Test previous works with satisfy()."""
        from pycsp3 import satisfy
        
        tasks = [IntervalVar(size=5, name=f"t{i}") for i in range(3)]
        
        satisfy(previous(tasks, tasks[0], tasks[1]))

    def test_satisfy_with_transition_matrix(self):
        """Test SeqNoOverlap with transition matrix works with satisfy()."""
        from pycsp3 import satisfy
        
        tasks = [IntervalVar(size=5, name=f"t{i}") for i in range(3)]
        seq = SequenceVar(intervals=tasks, types=[0, 1, 0], name="machine")
        matrix = [[0, 5], [3, 0]]
        
        satisfy(SeqNoOverlap(seq, transition_matrix=matrix))

    def test_satisfy_with_same_sequence(self):
        """Test same_sequence works with satisfy()."""
        from pycsp3 import satisfy
        
        tasks = [IntervalVar(size=5, name=f"t{i}") for i in range(3)]
        seq1 = SequenceVar(intervals=tasks, name="m1")
        seq2 = SequenceVar(intervals=tasks, name="m2")
        
        satisfy(same_sequence(seq1, seq2))

    def test_job_shop_pattern(self):
        """Test job shop scheduling pattern."""
        from pycsp3 import satisfy
        
        # 2 jobs, 3 operations each
        ops = [[IntervalVar(size=5, name=f"j{j}o{o}") for o in range(3)] for j in range(2)]
        
        # Operations of same job must be in order
        for j in range(2):
            for o in range(2):
                satisfy(before(ops[j], ops[j][o], ops[j][o + 1]))

    def test_combined_sequence_constraints(self):
        """Test combining multiple sequence constraints."""
        from pycsp3 import satisfy
        
        tasks = [IntervalVar(size=5, name=f"t{i}") for i in range(4)]
        seq = SequenceVar(intervals=tasks, name="machine")
        
        # t0 is first, t3 is last, t1 before t2
        satisfy(first(seq, tasks[0]))
        satisfy(last(seq, tasks[3]))
        satisfy(before(seq, tasks[1], tasks[2]))



"""
Cumulative constraints for resource modeling.

This module provides constraints on cumulative functions that integrate
with pycsp3's constraint system.

The key challenge is that pycsp3 doesn't have native cumulative functions,
so we decompose them into discrete time-indexed constraints or use the
Cumulative global constraint where applicable.
"""

from __future__ import annotations

from typing import TYPE_CHECKING, Sequence, Union

from pycsp3_scheduling.constraints._pycsp3 import length_value, presence_var, start_var
from pycsp3_scheduling.functions.cumul_functions import (
    CumulConstraint,
    CumulConstraintType,
    CumulExpr,
    CumulExprType,
    CumulFunction,
)

if TYPE_CHECKING:
    from pycsp3_scheduling.variables.interval import IntervalVar


def _get_node_builders():
    """Import and return pycsp3 Node building utilities."""
    from pycsp3.classes.nodes import Node, TypeNode

    return Node, TypeNode


def _build_end_expr(interval, Node, TypeNode):
    """Build end expression: start + length."""
    start = start_var(interval)
    length = length_value(interval)
    if isinstance(length, int) and length == 0:
        return start
    return Node.build(TypeNode.ADD, start, length)


def _is_simple_pulse_cumul(cumul: CumulFunction) -> bool:
    """Check if cumulative function is a simple sum of pulses with fixed heights."""
    for expr in cumul.expressions:
        if expr.expr_type == CumulExprType.NEG:
            # Check the negated expression
            if expr.operands and expr.operands[0].expr_type != CumulExprType.PULSE:
                return False
            if expr.operands and expr.operands[0].is_variable_height:
                return False
        elif expr.expr_type != CumulExprType.PULSE:
            return False
        elif expr.is_variable_height:
            return False
    return True


def _get_pulse_data(cumul: CumulFunction) -> tuple[list, list[int], list[int]]:
    """Extract intervals, heights, and lengths from a pulse-only cumulative."""
    intervals = []
    heights = []

    for expr in cumul.expressions:
        if expr.expr_type == CumulExprType.NEG:
            inner = expr.operands[0]
            intervals.append(inner.interval)
            heights.append(-(inner.height or inner.height_min))
        else:
            intervals.append(expr.interval)
            heights.append(expr.height or expr.height_min)

    return intervals, heights


def build_cumul_constraint(constraint: CumulConstraint) -> list:
    """
    Build pycsp3 constraints from a CumulConstraint.

    For simple pulse-based cumulative functions with capacity constraints,
    uses pycsp3's Cumulative global constraint. For more complex cases,
    decomposes into time-indexed constraints.

    Args:
        constraint: The cumulative constraint to build.

    Returns:
        List of pycsp3 constraint nodes or ECtr objects.
    """
    cumul = constraint.cumul

    # Handle simple capacity constraints on pulse cumulative functions
    if (
        constraint.constraint_type in (CumulConstraintType.LE, CumulConstraintType.RANGE)
        and _is_simple_pulse_cumul(cumul)
        and cumul.expressions
    ):
        return _build_cumulative_constraint(constraint)

    # For other cases, use decomposition
    return _build_decomposed_constraint(constraint)


def _build_cumulative_constraint(constraint: CumulConstraint) -> list:
    """
    Build using pycsp3's Cumulative global constraint.

    Cumulative(origins, lengths, heights) <= limit
    """
    from pycsp3 import Cumulative

    cumul = constraint.cumul
    intervals, heights = _get_pulse_data(cumul)

    # Filter out negative heights (not supported by standard Cumulative)
    if any(h < 0 for h in heights):
        return _build_decomposed_constraint(constraint)

    # Get pycsp3 variables
    origins = [start_var(iv) for iv in intervals]
    lengths = [length_value(iv) for iv in intervals]

    if constraint.constraint_type == CumulConstraintType.LE:
        limit = constraint.bound
    elif constraint.constraint_type == CumulConstraintType.RANGE:
        # Cumulative only supports upper bound directly
        # For lower bound, we'd need decomposition
        if constraint.min_bound > 0:
            return _build_decomposed_constraint(constraint)
        limit = constraint.max_bound
    else:
        return _build_decomposed_constraint(constraint)

    # Build the Cumulative constraint
    return [Cumulative(origins=origins, lengths=lengths, heights=heights) <= limit]


def _build_decomposed_constraint(constraint: CumulConstraint) -> list:
    """
    Build decomposed constraints for complex cumulative functions.

    This handles:
    - Variable heights
    - Step functions (step_at_start, step_at_end, step_at)
    - ALWAYS_IN constraints
    - GE/GT constraints

    For now, returns a placeholder as full decomposition is complex.
    """
    Node, TypeNode = _get_node_builders()
    constraints = []

    cumul = constraint.cumul

    # For GE constraints: can negate and use LE
    if constraint.constraint_type == CumulConstraintType.GE:
        # cumul >= bound  <=>  -cumul <= -bound
        # But we need to handle this properly with the actual expressions
        pass

    # For ALWAYS_IN: constrain during specific time range
    if constraint.constraint_type == CumulConstraintType.ALWAYS_IN:
        # This requires time-indexed reasoning
        # For fixed time ranges, we can reason about which intervals overlap
        pass

    # Default: return empty list (constraint not yet enforced at pycsp3 level)
    # The CumulConstraint object still exists for model introspection
    return constraints


def SeqCumulative(
    intervals: Sequence[IntervalVar],
    heights: Sequence[int],
    capacity: int,
) -> list:
    """
    Resource capacity constraint using pycsp3's Cumulative.

    Ensures that the sum of heights of intervals executing at any time
    does not exceed the capacity.

    This is named SeqCumulative to avoid clashing with pycsp3's native
    Cumulative constraint.

    Args:
        intervals: List of interval variables.
        heights: List of resource heights (demands) for each interval.
        capacity: Maximum capacity of the resource.

    Returns:
        List containing the Cumulative constraint.

    Raises:
        ValueError: If intervals and heights have different lengths.
        TypeError: If inputs have wrong types.

    Example:
        >>> tasks = [IntervalVar(size=5, name=f"t{i}") for i in range(3)]
        >>> demands = [2, 3, 1]
        >>> satisfy(SeqCumulative(tasks, demands, capacity=4))
    """
    from pycsp3_scheduling.variables.interval import IntervalVar
    from pycsp3 import Cumulative as Cumul

    if len(intervals) != len(heights):
        raise ValueError(
            f"intervals ({len(intervals)}) and heights ({len(heights)}) must have same length"
        )

    for i, iv in enumerate(intervals):
        if not isinstance(iv, IntervalVar):
            raise TypeError(f"intervals[{i}] must be an IntervalVar")

    for i, h in enumerate(heights):
        if not isinstance(h, int):
            raise TypeError(f"heights[{i}] must be an int")

    if not isinstance(capacity, int):
        raise TypeError(f"capacity must be an int, got {type(capacity).__name__}")

    origins = [start_var(iv) for iv in intervals]
    lengths = [length_value(iv) for iv in intervals]

    return [Cumul(origins=origins, lengths=lengths, heights=list(heights)) <= capacity]



"""
Grouping constraints for interval variables.

This module provides constraints that relate a main interval to a group of intervals:

1. **span(main, subtasks)**: Main interval spans all present subtasks
   - start(main) = min{start(i) : i present}
   - end(main) = max{end(i) : i present}

2. **alternative(main, alternatives, cardinality=1)**: Select exactly k alternatives
   - Exactly `cardinality` intervals from `alternatives` are present
   - Selected intervals have same start/end as main

3. **synchronize(main, intervals)**: All present intervals synchronize with main
   - All present intervals in array have same start/end as main

All constraints handle optional (absent) intervals correctly:
- When main is absent, all related intervals are absent
- When all subtasks are absent, main is absent (for span)
"""

from __future__ import annotations

from typing import TYPE_CHECKING, Sequence

from pycsp3_scheduling.constraints._pycsp3 import (
    length_value,
    presence_var,
    start_var,
)
from pycsp3_scheduling.variables.interval import IntervalVar

if TYPE_CHECKING:
    pass


def _validate_interval(interval: IntervalVar, name: str) -> None:
    """Validate that argument is an IntervalVar."""
    if not isinstance(interval, IntervalVar):
        raise TypeError(f"{name} must be an IntervalVar, got {type(interval).__name__}")


def _validate_interval_list(
    intervals: Sequence[IntervalVar], name: str
) -> list[IntervalVar]:
    """Validate and convert interval sequence."""
    if not isinstance(intervals, (list, tuple)):
        raise TypeError(f"{name} must be a list or tuple of IntervalVar")
    result = list(intervals)
    for i, interval in enumerate(result):
        if not isinstance(interval, IntervalVar):
            raise TypeError(
                f"{name}[{i}] must be an IntervalVar, got {type(interval).__name__}"
            )
    if len(result) == 0:
        raise ValueError(f"{name} cannot be empty")
    return result


def _get_node_builders():
    """Import and return pycsp3 Node building utilities."""
    from pycsp3.classes.nodes import Node, TypeNode

    return Node, TypeNode


def _build_end_expr(interval: IntervalVar, Node, TypeNode):
    """Build end expression: start + length."""
    start = start_var(interval)
    length = length_value(interval)
    if isinstance(length, int) and length == 0:
        return start
    return Node.build(TypeNode.ADD, start, length)


# =============================================================================
# Span Constraint
# =============================================================================


def span(main: IntervalVar, subtasks: Sequence[IntervalVar]) -> list:
    """
    Constrain main interval to span all present subtasks.

    The main interval covers exactly the time range occupied by its subtasks:
    - start(main) = min{start(i) : i present in subtasks}
    - end(main) = max{end(i) : i present in subtasks}

    Semantics for optional intervals:
    - If main is present, at least one subtask must be present
    - If main is absent, all subtasks must be absent
    - If all subtasks are absent, main must be absent

    Args:
        main: The spanning interval variable.
        subtasks: List of interval variables to be spanned.

    Returns:
        List of pycsp3 constraint nodes.

    Raises:
        TypeError: If main is not IntervalVar or subtasks contains non-IntervalVar.
        ValueError: If subtasks is empty.

    Example:
        >>> main_task = IntervalVar(name="project")
        >>> phases = [IntervalVar(size=10, name=f"phase_{i}") for i in range(3)]
        >>> satisfy(span(main_task, phases))  # project spans all phases
    """
    _validate_interval(main, "main")
    subtasks_list = _validate_interval_list(subtasks, "subtasks")
    Node, TypeNode = _get_node_builders()

    constraints = []
    main_start = start_var(main)
    main_end = _build_end_expr(main, Node, TypeNode)

    # Collect start and end expressions for subtasks
    subtask_starts = [start_var(t) for t in subtasks_list]
    subtask_ends = [_build_end_expr(t, Node, TypeNode) for t in subtasks_list]

    has_optional_main = main.optional
    has_optional_subtasks = any(t.optional for t in subtasks_list)

    if not has_optional_main and not has_optional_subtasks:
        # Simple case: all intervals are mandatory
        # start(main) = min(subtask starts)
        # end(main) = max(subtask ends)
        if len(subtasks_list) == 1:
            # Single subtask: direct equality
            constraints.append(Node.build(TypeNode.EQ, main_start, subtask_starts[0]))
            constraints.append(Node.build(TypeNode.EQ, main_end, subtask_ends[0]))
        else:
            min_start = Node.build(TypeNode.MIN, subtask_starts)
            max_end = Node.build(TypeNode.MAX, subtask_ends)
            constraints.append(Node.build(TypeNode.EQ, main_start, min_start))
            constraints.append(Node.build(TypeNode.EQ, main_end, max_end))
    else:
        # Complex case with optional intervals
        # We need conditional constraints based on presence

        main_presence = presence_var(main)
        subtask_presences = [presence_var(t) for t in subtasks_list]

        # For each subtask: if present, main must contain it
        for i, subtask in enumerate(subtasks_list):
            sub_start = subtask_starts[i]
            sub_end = subtask_ends[i]
            sub_presence = subtask_presences[i]

            if subtask.optional:
                # If subtask present: main_start <= sub_start AND sub_end <= main_end
                # Encoded as: (sub_presence == 0) OR (main_start <= sub_start)
                cond1 = Node.build(
                    TypeNode.OR,
                    Node.build(TypeNode.EQ, sub_presence, 0),
                    Node.build(TypeNode.LE, main_start, sub_start),
                )
                cond2 = Node.build(
                    TypeNode.OR,
                    Node.build(TypeNode.EQ, sub_presence, 0),
                    Node.build(TypeNode.LE, sub_end, main_end),
                )
                constraints.append(cond1)
                constraints.append(cond2)

                # If subtask present, main must be present
                if has_optional_main:
                    # sub_presence <= main_presence (if sub present, main present)
                    constraints.append(
                        Node.build(TypeNode.LE, sub_presence, main_presence)
                    )
            else:
                # Mandatory subtask: main must contain it
                constraints.append(Node.build(TypeNode.LE, main_start, sub_start))
                constraints.append(Node.build(TypeNode.LE, sub_end, main_end))

        if has_optional_main:
            # If main is present, at least one subtask must be present
            # main_presence <= sum(subtask_presences)
            # Equivalently: main_presence == 0 OR sum(subtask_presences) >= 1
            if has_optional_subtasks:
                sum_presences = Node.build(TypeNode.ADD, *subtask_presences)
                constraints.append(
                    Node.build(
                        TypeNode.OR,
                        Node.build(TypeNode.EQ, main_presence, 0),
                        Node.build(TypeNode.GE, sum_presences, 1),
                    )
                )

            # Main start/end should match the actual span when present
            # This is enforced by the containment constraints above plus tightness

    return constraints


# =============================================================================
# Alternative Constraint
# =============================================================================


def alternative(
    main: IntervalVar,
    alternatives: Sequence[IntervalVar],
    cardinality: int = 1,
) -> list:
    """
    Constrain exactly k alternatives to be selected and match the main interval.

    When main is present, exactly `cardinality` intervals from `alternatives`
    are present, and they have the same start and end times as main.

    Semantics:
    - If main is present: exactly `cardinality` alternatives are present
    - Selected alternatives have start(alt) == start(main) and end(alt) == end(main)
    - If main is absent: all alternatives are absent

    Args:
        main: The main interval variable.
        alternatives: List of alternative interval variables.
        cardinality: Number of alternatives to select (default 1).

    Returns:
        List of pycsp3 constraint nodes.

    Raises:
        TypeError: If main is not IntervalVar or alternatives contains non-IntervalVar.
        ValueError: If alternatives is empty or cardinality is invalid.

    Example:
        >>> main_op = IntervalVar(size=10, name="operation")
        >>> machines = [IntervalVar(size=10, optional=True, name=f"m{i}") for i in range(3)]
        >>> satisfy(alternative(main_op, machines))  # Assign to exactly one machine
    """
    _validate_interval(main, "main")
    alts_list = _validate_interval_list(alternatives, "alternatives")

    if not isinstance(cardinality, int) or cardinality < 1:
        raise ValueError(f"cardinality must be a positive integer, got {cardinality}")
    if cardinality > len(alts_list):
        raise ValueError(
            f"cardinality ({cardinality}) cannot exceed number of alternatives "
            f"({len(alts_list)})"
        )

    # All alternatives must be optional for alternative constraint to make sense
    for i, alt in enumerate(alts_list):
        if not alt.optional:
            raise ValueError(
                f"alternatives[{i}] ({alt.name}) must be optional for alternative constraint"
            )

    Node, TypeNode = _get_node_builders()
    constraints = []

    main_start = start_var(main)
    main_end = _build_end_expr(main, Node, TypeNode)
    main_presence = presence_var(main)

    alt_presences = [presence_var(alt) for alt in alts_list]

    # Sum of alternative presences equals cardinality when main is present
    # sum(alt_presences) == cardinality * main_presence
    if len(alt_presences) == 1:
        sum_presences = alt_presences[0]
    else:
        sum_presences = Node.build(TypeNode.ADD, *alt_presences)
    if main.optional:
        # sum(alt_presences) == cardinality * main_presence
        rhs = Node.build(TypeNode.MUL, cardinality, main_presence)
        constraints.append(Node.build(TypeNode.EQ, sum_presences, rhs))
    else:
        # Main is mandatory, so exactly cardinality alternatives must be present
        constraints.append(Node.build(TypeNode.EQ, sum_presences, cardinality))

    # Each selected alternative must match main's timing
    for i, alt in enumerate(alts_list):
        alt_start = start_var(alt)
        alt_end = _build_end_expr(alt, Node, TypeNode)
        alt_presence = alt_presences[i]

        # If alternative is present, it must match main's start
        # (alt_presence == 0) OR (alt_start == main_start)
        start_match = Node.build(
            TypeNode.OR,
            Node.build(TypeNode.EQ, alt_presence, 0),
            Node.build(TypeNode.EQ, alt_start, main_start),
        )
        constraints.append(start_match)

        # If alternative is present, it must match main's end
        # (alt_presence == 0) OR (alt_end == main_end)
        end_match = Node.build(
            TypeNode.OR,
            Node.build(TypeNode.EQ, alt_presence, 0),
            Node.build(TypeNode.EQ, alt_end, main_end),
        )
        constraints.append(end_match)

    return constraints


# =============================================================================
# Synchronize Constraint
# =============================================================================


def synchronize(main: IntervalVar, intervals: Sequence[IntervalVar]) -> list:
    """
    Constrain all present intervals to synchronize with the main interval.

    All present intervals in the array have the same start and end times as main.

    Semantics:
    - If main is present: present intervals have same start/end as main
    - If main is absent: all intervals in array are absent
    - Intervals in array can be independently present or absent

    Args:
        main: The main interval variable.
        intervals: List of interval variables to synchronize.

    Returns:
        List of pycsp3 constraint nodes.

    Raises:
        TypeError: If main is not IntervalVar or intervals contains non-IntervalVar.
        ValueError: If intervals is empty.

    Example:
        >>> main = IntervalVar(size=10, name="meeting")
        >>> attendees = [IntervalVar(size=10, optional=True, name=f"person_{i}") for i in range(5)]
        >>> satisfy(synchronize(main, attendees))  # Present attendees sync with meeting
    """
    _validate_interval(main, "main")
    intervals_list = _validate_interval_list(intervals, "intervals")
    Node, TypeNode = _get_node_builders()

    constraints = []
    main_start = start_var(main)
    main_end = _build_end_expr(main, Node, TypeNode)
    main_presence = presence_var(main)

    for i, interval in enumerate(intervals_list):
        int_start = start_var(interval)
        int_end = _build_end_expr(interval, Node, TypeNode)
        int_presence = presence_var(interval)

        if interval.optional:
            # If main is absent, interval must be absent
            # main_presence >= int_presence (if main absent, interval absent)
            if main.optional:
                constraints.append(
                    Node.build(TypeNode.GE, main_presence, int_presence)
                )
            else:
                # Main is mandatory, intervals can be present or absent
                pass

            # If interval is present, it must match main's timing
            # (int_presence == 0) OR (int_start == main_start)
            start_match = Node.build(
                TypeNode.OR,
                Node.build(TypeNode.EQ, int_presence, 0),
                Node.build(TypeNode.EQ, int_start, main_start),
            )
            constraints.append(start_match)

            # (int_presence == 0) OR (int_end == main_end)
            end_match = Node.build(
                TypeNode.OR,
                Node.build(TypeNode.EQ, int_presence, 0),
                Node.build(TypeNode.EQ, int_end, main_end),
            )
            constraints.append(end_match)
        else:
            # Mandatory interval: must always match main
            constraints.append(Node.build(TypeNode.EQ, int_start, main_start))
            constraints.append(Node.build(TypeNode.EQ, int_end, main_end))

    return constraints



"""
Flexible Job Shop Scheduling

This model uses pycsp3-scheduling's IntervalVar with optional intervals
and alternative constraints to model the Flexible Job Shop problem.

Each task can be executed on one of several alternative machines,
and the goal is to minimize the makespan.

## Data Example
  easy01.json

## Model
  variables: IntervalVar (optional)
  constraints: end_before_start, alternative, SeqCumulative

## Execution
  python FlexibleJobshop.py -data=<datafile.json>

## Links
  - https://www.minizinc.org/challenge/2013/results/
  - https://github.com/xcsp3team/PyCSP3-models/blob/main/realistic/FlexibleJobshop/FlexibleJobshop.py

## Tags
  realistic, scheduling, mzn13
"""

import json
import os

from pycsp3 import Var, satisfy, minimize, solve, SAT, OPTIMUM, value
from pycsp3_scheduling import (
    IntervalVar,
    end_before_start,
    end_time,
    alternative,
    SeqCumulative,
    interval_value,
)

# Load data
_data_dir = os.path.join(os.path.dirname(__file__), "data")
with open(os.path.join(_data_dir, "easy01.json")) as f:
    _data = json.load(f)

nMachines = _data["nMachines"]
tasks = _data["tasks"]
options = _data["optionalTasks"]
option_machines = _data["machines"]
option_durations = _data["durations"]

nJobs, nTasks, nOptions = len(tasks), len(options), len(option_machines)
J, T, O, M = range(nJobs), range(nTasks), range(nOptions), range(nMachines)

# Compute auxiliary information
siblings = [next(tasks[j] for j in J if t in tasks[j]) for t in T]
minDurations = [min(option_durations[options[t]]) for t in T]
maxDurations = [max(option_durations[options[t]]) for t in T]
minStarts = [sum(minDurations[k] for k in siblings[t] if k < t) for t in T]
horizon = sum(option_durations)
maxStarts = [horizon - sum(minDurations[k] for k in siblings[t] if k >= t) for t in T]

taskForOption = [next(t for t in T if o in options[t]) for o in O]

# task_intervals[t] is the main interval for task t
task_intervals = [
    IntervalVar(
        start=(minStarts[t], maxStarts[t]),
        size=(minDurations[t], maxDurations[t]),
        name=f"task_{t}",
    )
    for t in T
]

# opt_intervals[o] is the optional interval for option o
opt_intervals = [
    IntervalVar(
        start=(minStarts[taskForOption[o]], maxStarts[taskForOption[o]]),
        size=option_durations[o],
        optional=True,
        name=f"opt_{o}",
    )
    for o in O
]

# makespan is the objective
makespan = Var(dom=range(horizon + 1))

satisfy(
    # precedence: tasks within a job must be sequential
    [
        end_before_start(task_intervals[t], task_intervals[t + 1])
        for j in J
        for t in tasks[j][:-1]
    ],
    # alternative: each task is performed by exactly one option
    [alternative(task_intervals[t], [opt_intervals[o] for o in options[t]]) for t in T],
    # cumulative: at most one operation per machine at a time
    [
        SeqCumulative(
            [opt_intervals[o] for o in O if option_machines[o] == m],
            heights=[1 for o in O if option_machines[o] == m],
            capacity=1,
        )
        for m in M
    ],
    # makespan
    [makespan >= end_time(task_intervals[tasks[j][-1]]) for j in J],
)

minimize(makespan)

# --- Solution output ---
if __name__ == "__main__":
    if solve() in (SAT, OPTIMUM):
        print(f"Makespan: {value(makespan)}")
        for j in J:
            print(f"Job {j}:")
            for t in tasks[j]:
                v = interval_value(task_intervals[t])
                # Find which option was selected
                selected_opt = None
                for o in options[t]:
                    ov = interval_value(opt_intervals[o])
                    if ov is not None and ov["present"]:
                        selected_opt = o
                        break
                machine = option_machines[selected_opt] if selected_opt is not None else "?"
                print(f"  Task {t} on M{machine}: [{v['start']}, {v['end']})")



from pycsp3 import *
from pycsp3_scheduling import *

# Data
durations = [3, 2, 5, 4, 2]
demands = [[2, 1], [1, 2], [3, 0], [2, 1], [1, 3]]
capacities = [4, 3]
precedences = [(0, 2), (1, 3), (2, 4)]

# Interval variables
tasks = [IntervalVar(size=durations[i], name=f"task_{i}")
         for i in range(len(durations))]

# Precedence constraints
satisfy(end_before_start(tasks[p], tasks[s]) for p, s in precedences)

# Cumulative resource constraints
for r in range(len(capacities)):
    resource = sum(pulse(tasks[i], demands[i][r])
                   for i in range(len(tasks)) if demands[i][r] > 0)
    satisfy(resource <= capacities[r])

# Minimize makespan
minimize(Maximum(end_time(t) for t in tasks))

if solve() in [SAT, OPTIMUM]:
    for t in tasks:
        v = interval_value(t)
        print(f"{t.name}: start={v['start']}, end={v['end']}")
else:
    print("No solution found")


"""
Flexible Job Shop Scheduling Problem (FJSP)

In the flexible job shop, each operation can be processed on any machine
from a given set of eligible machines (with potentially different durations).

This example models machine selection using optional interval variables.
Each operation has multiple optional intervals (one per eligible machine),
and exactly one must be selected.
"""
from pycsp3 import *
from pycsp3_scheduling import *

# Data: jobs with operations, each operation has (machine, duration) alternatives
# Job format: [[(m1, d1), (m2, d2), ...], ...]  for each operation
jobs_data = [
    # Job 0: 2 operations
    [[(0, 3), (1, 4)],           # op0 can run on machine 0 (3) or machine 1 (4)
     [(1, 2), (2, 3)]],          # op1 can run on machine 1 (2) or machine 2 (3)
    # Job 1: 2 operations  
    [[(0, 2), (2, 4)],           # op0 can run on machine 0 (2) or machine 2 (4)
     [(0, 3), (1, 2)]],          # op1 can run on machine 0 (3) or machine 1 (2)
    # Job 2: 2 operations
    [[(0, 2), (1, 3)],           # op0 can run on machine 0 (2) or machine 1 (3)
     [(1, 2), (2, 4)]],          # op1 can run on machine 1 (2) or machine 2 (4)
]

n_jobs = len(jobs_data)
n_machines = 3

# Create optional interval variables for each (operation, machine) combination
# ops[j][o] = list of optional intervals, one per eligible machine
ops = [[
    [IntervalVar(size=d, optional=True, name=f"j{j}_o{o}_m{m}")
     for m, d in jobs_data[j][o]]
    for o in range(len(jobs_data[j]))]
    for j in range(n_jobs)]

# Store presence and start variables for use in constraints
# (We need these for the Sum and precedence constraints)
presence = [[
    [presence_time(ops[j][o][idx]) for idx in range(len(jobs_data[j][o]))]
    for o in range(len(jobs_data[j]))]
    for j in range(n_jobs)]

starts = [[
    [start_time(ops[j][o][idx]) for idx in range(len(jobs_data[j][o]))]
    for o in range(len(jobs_data[j]))]
    for j in range(n_jobs)]

# Exactly one machine selected per operation
satisfy(
    Sum(presence[j][o]) == 1
    for j in range(n_jobs)
    for o in range(len(jobs_data[j]))
)

# Precedence within jobs: each selected op must finish before next op starts
for j in range(n_jobs):
    for o in range(len(jobs_data[j]) - 1):
        for idx1, (m1, d1) in enumerate(jobs_data[j][o]):
            for idx2, (m2, d2) in enumerate(jobs_data[j][o + 1]):
                # If both are selected, precedence must hold
                satisfy(
                    (presence[j][o][idx1] == 0) | (presence[j][o + 1][idx2] == 0) |
                    (starts[j][o][idx1] + d1 <= starts[j][o + 1][idx2])
                )

# No overlap on each machine
for m in range(n_machines):
    machine_ops = []
    machine_durations = []
    for j in range(n_jobs):
        for o in range(len(jobs_data[j])):
            for idx, (machine, duration) in enumerate(jobs_data[j][o]):
                if machine == m:
                    machine_ops.append(ops[j][o][idx])
                    machine_durations.append(duration)
    if machine_ops:
        seq = SequenceVar(intervals=machine_ops, name=f"machine_{m}")
        satisfy(SeqNoOverlap(seq))

# Minimize makespan: max end time of last operations
last_op_ends = []
for j in range(n_jobs):
    o = len(jobs_data[j]) - 1  # Last operation
    for idx, (m, d) in enumerate(jobs_data[j][o]):
        last_op_ends.append(starts[j][o][idx] + d * presence[j][o][idx])

minimize(Maximum(last_op_ends))

# Solve
if solve() in [SAT, OPTIMUM]:
    print("Solution found!")
    print()
    makespan = 0
    for j in range(n_jobs):
        print(f"Job {j}:")
        for o in range(len(jobs_data[j])):
            for idx, (m, d) in enumerate(jobs_data[j][o]):
                v = interval_value(ops[j][o][idx])
                if v is not None:  # This alternative was selected
                    makespan = max(makespan, v['end'])
                    print(f"  Op {o}: machine {m}, [{v['start']}-{v['end']}]")
                    break
    print()
    print(f"Makespan: {makespan}")
else:
    print("No solution found")



"""
Scheduling function types.

This module provides:
- CumulFunction: Cumulative function for resource consumption
- pulse, step_at, step_at_start, step_at_end: Elementary cumul functions
- cumul_range, always_in: Cumulative constraints
- height_at_start, height_at_end: Cumulative accessors
- StateFunction: State function for resource states
- TransitionMatrix: Transition costs/times between states
- always_equal, always_constant, always_no_state: State constraints
- StepFunction: Piecewise step function for forbidden regions (future)
"""

from __future__ import annotations

from pycsp3_scheduling.functions.cumul_functions import (
    CumulExpr,
    CumulExprType,
    CumulFunction,
    CumulConstraint,
    CumulConstraintType,
    CumulHeightExpr,
    CumulHeightType,
    always_in as cumul_always_in,
    cumul_range,
    height_at_end,
    height_at_start,
    pulse,
    step_at,
    step_at_end,
    step_at_start,
    clear_cumul_registry,
    get_registered_cumuls,
    register_cumul,
)

from pycsp3_scheduling.functions.state_functions import (
    StateFunction,
    StateConstraint,
    StateConstraintType,
    TransitionMatrix,
    always_constant,
    always_equal,
    always_in as state_always_in,
    always_no_state,
    clear_state_function_registry,
    get_registered_state_functions,
)

# Re-export always_in with smart dispatch
def always_in(resource, interval_or_range, min_val, max_val=None, **kwargs):
    """
    Constrain a cumulative or state function within bounds.
    
    For CumulFunction: always_in(cumul, interval_or_range, min, max)
    For StateFunction: always_in(state_func, interval, min, max)
    """
    if isinstance(resource, CumulFunction):
        return cumul_always_in(resource, interval_or_range, min_val, max_val)
    elif isinstance(resource, StateFunction):
        return state_always_in(resource, interval_or_range, min_val, max_val, **kwargs)
    else:
        raise TypeError(
            f"always_in expects CumulFunction or StateFunction, got {type(resource).__name__}"
        )


# To be implemented:
# from pycsp3_scheduling.functions.step_function import StepFunction

__all__ = [
    # Cumulative function types
    "CumulExpr",
    "CumulExprType",
    "CumulFunction",
    "CumulConstraint",
    "CumulConstraintType",
    "CumulHeightExpr",
    "CumulHeightType",
    # Elementary cumul functions
    "pulse",
    "step_at",
    "step_at_start",
    "step_at_end",
    # Cumul constraints
    "cumul_range",
    # Cumul accessors
    "height_at_start",
    "height_at_end",
    # Cumul registry
    "register_cumul",
    "get_registered_cumuls",
    "clear_cumul_registry",
    # State function types
    "StateFunction",
    "StateConstraint",
    "StateConstraintType",
    "TransitionMatrix",
    # State constraints
    "always_equal",
    "always_constant",
    "always_no_state",
    # State registry
    "get_registered_state_functions",
    "clear_state_function_registry",
    # Shared
    "always_in",
    # Future
    # "StepFunction",
]



"""Tests for scheduling constraints."""

import pytest

pycsp3 = pytest.importorskip("pycsp3")

from pycsp3.classes.entities import ECtr, clear as clear_pycsp3
from pycsp3.classes.main.constraints import ConstraintNoOverlap
from pycsp3.classes.nodes import Node, TypeNode

from pycsp3_scheduling.constraints import (
    SeqNoOverlap,
    end_at_end,
    end_at_start,
    end_before_end,
    end_before_start,
    start_at_end,
    start_at_start,
    start_before_end,
    start_before_start,
)
from pycsp3_scheduling.constraints._pycsp3 import clear_pycsp3_cache
from pycsp3_scheduling.variables import (
    IntervalVar,
    SequenceVar,
    clear_interval_registry,
    clear_sequence_registry,
)


@pytest.fixture(autouse=True)
def reset_state():
    """Reset registries and caches before each test."""
    clear_pycsp3()
    clear_interval_registry()
    clear_sequence_registry()
    clear_pycsp3_cache()
    yield
    clear_pycsp3()
    clear_interval_registry()
    clear_sequence_registry()
    clear_pycsp3_cache()


class TestSeqNoOverlap:
    """Tests for SeqNoOverlap constraint."""

    def test_sequence_var(self):
        """SeqNoOverlap accepts a SequenceVar."""
        task1 = IntervalVar(size=3, name="t1")
        task2 = IntervalVar(size=2, name="t2")
        seq = SequenceVar(intervals=[task1, task2], name="machine")

        ctr = SeqNoOverlap(seq)

        assert isinstance(ctr, ECtr)
        assert isinstance(ctr.constraint, ConstraintNoOverlap)

    def test_interval_list(self):
        """SeqNoOverlap accepts a list of IntervalVar."""
        task1 = IntervalVar(size=3, name="t1")
        task2 = IntervalVar(size=2, name="t2")

        ctr = SeqNoOverlap([task1, task2])

        assert isinstance(ctr, ECtr)
        assert isinstance(ctr.constraint, ConstraintNoOverlap)

    def test_transition_matrix_now_supported(self):
        """transition_matrix is now supported with typed SequenceVar."""
        tasks = [IntervalVar(size=3, name="t1"), IntervalVar(size=2, name="t2")]
        seq = SequenceVar(intervals=tasks, types=[0, 1], name="machine")

        # Should not raise - now supported
        result = SeqNoOverlap(seq, transition_matrix=[[0, 5], [3, 0]])
        assert isinstance(result, list)

    def test_transition_matrix_requires_types(self):
        """transition_matrix requires SequenceVar with types."""
        task = IntervalVar(size=3, name="t1")
        seq = SequenceVar(intervals=[task], name="machine")

        with pytest.raises(ValueError, match="requires a SequenceVar with types"):
            SeqNoOverlap(seq, transition_matrix=[[0]])

    def test_is_direct_parameter(self):
        """is_direct parameter is accepted (no special behavior yet)."""
        tasks = [IntervalVar(size=3, name="t1"), IntervalVar(size=2, name="t2")]
        seq = SequenceVar(intervals=tasks, types=[0, 1], name="machine")

        # is_direct can be used with transition matrix
        result = SeqNoOverlap(seq, transition_matrix=[[0, 5], [3, 0]], is_direct=True)
        assert isinstance(result, list)

    def test_invalid_sequence_type(self):
        """SeqNoOverlap rejects non-sequences."""
        with pytest.raises(TypeError, match="SequenceVar"):
            SeqNoOverlap(123)

    def test_invalid_interval_element(self):
        """SeqNoOverlap rejects non-IntervalVar elements."""
        with pytest.raises(TypeError, match="IntervalVar"):
            SeqNoOverlap([1, 2])

    def test_optional_interval_supported(self):
        """Optional intervals are now supported."""
        task = IntervalVar(size=3, optional=True, name="t1")
        task2 = IntervalVar(size=2, optional=True, name="t2")
        seq = SequenceVar(intervals=[task, task2], name="machine")

        # Should not raise - optional intervals now supported
        ctr = SeqNoOverlap(seq)
        assert isinstance(ctr, ECtr)


# =============================================================================
# Exact Timing Constraints (Equality)
# =============================================================================


class TestStartAtStart:
    """Tests for start_at_start constraint."""

    def test_basic_constraint(self):
        """start_at_start returns an equality node."""
        task1 = IntervalVar(size=3, name="t1")
        task2 = IntervalVar(size=2, name="t2")

        expr = start_at_start(task1, task2)

        assert isinstance(expr, Node)
        assert expr.type == TypeNode.EQ
        expr_str = str(expr)
        assert "iv_s_0" in expr_str
        assert "iv_s_1" in expr_str
        assert "eq(" in expr_str

    def test_with_delay(self):
        """start_at_start with delay adds to rhs."""
        task1 = IntervalVar(size=3, name="t1")
        task2 = IntervalVar(size=2, name="t2")

        expr = start_at_start(task1, task2, delay=5)

        assert isinstance(expr, Node)
        assert expr.type == TypeNode.EQ
        expr_str = str(expr)
        assert "5" in expr_str
        assert "add(" in expr_str

    def test_invalid_interval_type(self):
        """start_at_start rejects non-IntervalVar inputs."""
        task = IntervalVar(size=3, name="t1")

        with pytest.raises(TypeError, match="IntervalVar"):
            start_at_start("t0", task)

    def test_invalid_delay_type(self):
        """start_at_start rejects non-int delay."""
        task1 = IntervalVar(size=3, name="t1")
        task2 = IntervalVar(size=2, name="t2")

        with pytest.raises(TypeError, match="delay must be an int"):
            start_at_start(task1, task2, delay="5")


class TestStartAtEnd:
    """Tests for start_at_end constraint."""

    def test_basic_constraint(self):
        """start_at_end returns an equality node with length."""
        task1 = IntervalVar(size=3, name="t1")
        task2 = IntervalVar(size=2, name="t2")

        expr = start_at_end(task1, task2)

        assert isinstance(expr, Node)
        assert expr.type == TypeNode.EQ
        expr_str = str(expr)
        # Should have start vars and duration
        assert "iv_s_0" in expr_str
        assert "iv_s_1" in expr_str
        assert "3" in expr_str  # Fixed length of task1

    def test_with_delay(self):
        """start_at_end with delay adds to rhs."""
        task1 = IntervalVar(size=3, name="t1")
        task2 = IntervalVar(size=2, name="t2")

        expr = start_at_end(task1, task2, delay=2)

        assert isinstance(expr, Node)
        assert expr.type == TypeNode.EQ
        expr_str = str(expr)
        # delay (2) + length (3) = 5 in the expression
        assert "5" in expr_str

    def test_variable_length(self):
        """start_at_end works with variable-length intervals."""
        task1 = IntervalVar(size=(3, 10), name="t1")
        task2 = IntervalVar(size=2, name="t2")

        expr = start_at_end(task1, task2)

        assert isinstance(expr, Node)
        expr_str = str(expr)
        assert "iv_l_0" in expr_str  # Variable length for task1

    def test_invalid_interval_type(self):
        """start_at_end rejects non-IntervalVar inputs."""
        task = IntervalVar(size=3, name="t1")

        with pytest.raises(TypeError, match="IntervalVar"):
            start_at_end(task, "t2")


class TestEndAtStart:
    """Tests for end_at_start constraint."""

    def test_basic_constraint(self):
        """end_at_start returns an equality node."""
        task1 = IntervalVar(size=3, name="t1")
        task2 = IntervalVar(size=2, name="t2")

        expr = end_at_start(task1, task2)

        assert isinstance(expr, Node)
        assert expr.type == TypeNode.EQ
        expr_str = str(expr)
        assert "iv_s_0" in expr_str
        assert "iv_s_1" in expr_str

    def test_with_delay(self):
        """end_at_start with delay adds to rhs."""
        task1 = IntervalVar(size=3, name="t1")
        task2 = IntervalVar(size=2, name="t2")

        expr = end_at_start(task1, task2, delay=4)

        assert isinstance(expr, Node)
        expr_str = str(expr)
        assert "4" in expr_str

    def test_invalid_interval_type(self):
        """end_at_start rejects non-IntervalVar inputs."""
        task = IntervalVar(size=3, name="t1")

        with pytest.raises(TypeError, match="IntervalVar"):
            end_at_start(task, 123)


class TestEndAtEnd:
    """Tests for end_at_end constraint."""

    def test_basic_constraint(self):
        """end_at_end returns an equality node."""
        task1 = IntervalVar(size=3, name="t1")
        task2 = IntervalVar(size=2, name="t2")

        expr = end_at_end(task1, task2)

        assert isinstance(expr, Node)
        assert expr.type == TypeNode.EQ
        expr_str = str(expr)
        # Should have both start vars and both lengths
        assert "iv_s_0" in expr_str
        assert "iv_s_1" in expr_str
        assert "3" in expr_str  # task1 length
        assert "2" in expr_str  # task2 length

    def test_with_delay(self):
        """end_at_end with delay adds to rhs."""
        task1 = IntervalVar(size=3, name="t1")
        task2 = IntervalVar(size=2, name="t2")

        expr = end_at_end(task1, task2, delay=7)

        assert isinstance(expr, Node)
        assert expr.type == TypeNode.EQ
        expr_str = str(expr)
        # length(a)=3 + delay=7 = 10 in the expression
        assert "10" in expr_str

    def test_variable_lengths(self):
        """end_at_end works with variable-length intervals."""
        task1 = IntervalVar(size=(3, 10), name="t1")
        task2 = IntervalVar(size=(2, 8), name="t2")

        expr = end_at_end(task1, task2)

        assert isinstance(expr, Node)
        expr_str = str(expr)
        assert "iv_l_0" in expr_str  # Variable length for task1
        assert "iv_l_1" in expr_str  # Variable length for task2

    def test_invalid_interval_type(self):
        """end_at_end rejects non-IntervalVar inputs."""
        task = IntervalVar(size=3, name="t1")

        with pytest.raises(TypeError, match="IntervalVar"):
            end_at_end(None, task)


# =============================================================================
# Before Constraints (Inequality)
# =============================================================================


class TestStartBeforeStart:
    """Tests for start_before_start constraint."""

    def test_basic_constraint(self):
        """start_before_start returns an inequality node."""
        task1 = IntervalVar(size=3, name="t1")
        task2 = IntervalVar(size=2, name="t2")

        expr = start_before_start(task1, task2)

        assert isinstance(expr, Node)
        assert expr.type == TypeNode.LE
        expr_str = str(expr)
        assert "iv_s_0" in expr_str
        assert "iv_s_1" in expr_str
        assert "le(" in expr_str

    def test_with_delay(self):
        """start_before_start with delay."""
        task1 = IntervalVar(size=3, name="t1")
        task2 = IntervalVar(size=2, name="t2")

        expr = start_before_start(task1, task2, delay=3)

        assert isinstance(expr, Node)
        expr_str = str(expr)
        assert "3" in expr_str
        assert "add(" in expr_str

    def test_invalid_interval_type(self):
        """start_before_start rejects non-IntervalVar inputs."""
        task = IntervalVar(size=3, name="t1")

        with pytest.raises(TypeError, match="IntervalVar"):
            start_before_start("invalid", task)


class TestStartBeforeEnd:
    """Tests for start_before_end constraint."""

    def test_basic_constraint(self):
        """start_before_end returns an inequality node."""
        task1 = IntervalVar(size=3, name="t1")
        task2 = IntervalVar(size=2, name="t2")

        expr = start_before_end(task1, task2)

        assert isinstance(expr, Node)
        assert expr.type == TypeNode.LE
        expr_str = str(expr)
        assert "iv_s_0" in expr_str
        assert "iv_s_1" in expr_str
        assert "2" in expr_str  # task2 length

    def test_with_delay(self):
        """start_before_end with delay."""
        task1 = IntervalVar(size=3, name="t1")
        task2 = IntervalVar(size=2, name="t2")

        expr = start_before_end(task1, task2, delay=5)

        assert isinstance(expr, Node)
        expr_str = str(expr)
        assert "5" in expr_str

    def test_invalid_interval_type(self):
        """start_before_end rejects non-IntervalVar inputs."""
        task = IntervalVar(size=3, name="t1")

        with pytest.raises(TypeError, match="IntervalVar"):
            start_before_end(task, [])


class TestEndBeforeStart:
    """Tests for end_before_start constraint."""

    def test_basic_constraint(self):
        """end_before_start returns a comparison node."""
        task1 = IntervalVar(size=3, name="t1")
        task2 = IntervalVar(size=2, name="t2")

        expr = end_before_start(task1, task2, delay=1)

        assert isinstance(expr, Node)
        assert expr.type == TypeNode.LE
        expr_str = str(expr)
        assert "iv_s_0" in expr_str
        assert "iv_s_1" in expr_str
        assert "le(" in expr_str

    def test_no_delay(self):
        """end_before_start without delay (classic precedence)."""
        task1 = IntervalVar(size=3, name="t1")
        task2 = IntervalVar(size=2, name="t2")

        expr = end_before_start(task1, task2)

        assert isinstance(expr, Node)
        assert expr.type == TypeNode.LE
        expr_str = str(expr)
        assert "3" in expr_str  # task1 length

    def test_invalid_delay_type(self):
        """end_before_start rejects non-int delay."""
        task1 = IntervalVar(size=3, name="t1")
        task2 = IntervalVar(size=2, name="t2")

        with pytest.raises(TypeError, match="delay must be an int"):
            end_before_start(task1, task2, delay="1")

    def test_invalid_interval_type(self):
        """end_before_start rejects non-IntervalVar inputs."""
        task = IntervalVar(size=3, name="t1")

        with pytest.raises(TypeError, match="IntervalVar"):
            end_before_start("t0", task)


class TestEndBeforeEnd:
    """Tests for end_before_end constraint."""

    def test_basic_constraint(self):
        """end_before_end returns an inequality node."""
        task1 = IntervalVar(size=3, name="t1")
        task2 = IntervalVar(size=2, name="t2")

        expr = end_before_end(task1, task2)

        assert isinstance(expr, Node)
        assert expr.type == TypeNode.LE
        expr_str = str(expr)
        assert "iv_s_0" in expr_str
        assert "iv_s_1" in expr_str
        assert "3" in expr_str  # task1 length
        assert "2" in expr_str  # task2 length

    def test_with_delay(self):
        """end_before_end with delay."""
        task1 = IntervalVar(size=3, name="t1")
        task2 = IntervalVar(size=2, name="t2")

        expr = end_before_end(task1, task2, delay=4)

        assert isinstance(expr, Node)
        assert expr.type == TypeNode.LE
        expr_str = str(expr)
        # length(a)=3 + delay=4 = 7 in the expression
        assert "7" in expr_str

    def test_variable_lengths(self):
        """end_before_end works with variable-length intervals."""
        task1 = IntervalVar(size=(3, 10), name="t1")
        task2 = IntervalVar(size=(2, 8), name="t2")

        expr = end_before_end(task1, task2)

        assert isinstance(expr, Node)
        expr_str = str(expr)
        assert "iv_l_0" in expr_str
        assert "iv_l_1" in expr_str

    def test_invalid_interval_type(self):
        """end_before_end rejects non-IntervalVar inputs."""
        task = IntervalVar(size=3, name="t1")

        with pytest.raises(TypeError, match="IntervalVar"):
            end_before_end(task, {"not": "interval"})


# =============================================================================
# Integration Tests
# =============================================================================


class TestPrecedenceIntegration:
    """Integration tests for precedence constraints with pycsp3."""

    def test_satisfy_with_precedence(self):
        """Test that constraints can be used with satisfy()."""
        from pycsp3 import satisfy

        task1 = IntervalVar(size=3, name="t1")
        task2 = IntervalVar(size=2, name="t2")

        # Should not raise
        satisfy(end_before_start(task1, task2))

    def test_satisfy_with_multiple_constraints(self):
        """Test multiple precedence constraints together."""
        from pycsp3 import satisfy

        task1 = IntervalVar(size=3, name="t1")
        task2 = IntervalVar(size=2, name="t2")
        task3 = IntervalVar(size=4, name="t3")

        # Should not raise
        satisfy(
            end_before_start(task1, task2),
            start_at_start(task2, task3),
        )

    def test_satisfy_with_generator(self):
        """Test precedence constraints with generator expression."""
        from pycsp3 import satisfy

        tasks = [IntervalVar(size=i + 1, name=f"t{i}") for i in range(3)]

        # Chain of precedence constraints
        satisfy(
            end_before_start(tasks[i], tasks[i + 1]) for i in range(len(tasks) - 1)
        )

    def test_constraint_with_regular_pycsp3_var(self):
        """Test that interval vars coexist with regular pycsp3 variables."""
        from pycsp3 import Var, satisfy

        # Regular pycsp3 variable
        x = Var(dom=range(10), id="x")

        # Interval variable
        task = IntervalVar(size=3, name="task")

        # Get pycsp3 var from interval
        from pycsp3_scheduling.interop import start_time

        start = start_time(task)

        # Constraint mixing both
        satisfy(start <= x)

    def test_exact_and_before_constraints_together(self):
        """Test mixing exact timing and before constraints."""
        from pycsp3 import satisfy

        task1 = IntervalVar(size=3, name="t1")
        task2 = IntervalVar(size=2, name="t2")
        task3 = IntervalVar(size=4, name="t3")

        # task1 and task2 start together, task3 starts after task2 ends
        satisfy(
            start_at_start(task1, task2),
            end_before_start(task2, task3),
        )



"""
Open-shop Scheduling

This model uses pycsp3's native NoOverlap constraint for the Open Shop
Scheduling problem, since open-shop requires variable durations that depend
on machine assignment (not a fixed size per operation).

In an open shop, each job has operations on each machine, but the order
of machines for each job is flexible (to be determined by the solver).

## Data Example
  GP-os-01.json

## Model
  variables: Var (integer)
  constraints: NoOverlap, AllDifferent, Table

## Execution
  python SchedulingOS.py -data=<datafile.json>
  python SchedulingOS.py  (uses default data)

## Links
  - https://en.wikipedia.org/wiki/Open-shop_scheduling
  - https://github.com/xcsp3team/PyCSP3-models/blob/main/realistic/SchedulingOS/SchedulingOS.py

## Tags
  realistic, scheduling, xcsp25
"""

import json
import os

from pycsp3 import (
    Var, VarArray, satisfy, minimize, solve, SAT, OPTIMUM, value,
    AllDifferent, Table, NoOverlap
)

# Load data
_data_dir = os.path.join(os.path.dirname(__file__), "data")
with open(os.path.join(_data_dir, "GP-os-01.json")) as f:
    durations = json.load(f)

horizon = sum(sum(t) for t in durations) + 1

n, m = len(durations), len(durations[0])
N, M = range(n), range(m)

# s[i][j] is the start time of the jth operation of job i
s = VarArray(size=[n, m], dom=range(horizon))

# mc[i][j] is the machine used for the jth operation of job i
mc = VarArray(size=[n, m], dom=range(m))

# d[i][j] is the duration of the jth operation of job i
d = VarArray(size=[n, m], dom=lambda i, j: durations[i])

# sd[i][k] is the start time when machine k is used for job i
sd = VarArray(size=[n, m], dom=range(horizon))

# makespan is the objective
makespan = Var(dom=range(horizon + 1))

satisfy(
    # operations must be non-overlapping within each job (sequential)
    [s[i][j] + d[i][j] <= s[i][j + 1] for i in N for j in range(m - 1)],

    # each machine must be used exactly once per job
    [AllDifferent(mc[i]) for i in N],

    # link machine choice to duration via Table constraint
    [
        Table(
            scope=(mc[i][j], d[i][j]),
            supports=enumerate(durations[i]),
        )
        for j in M
        for i in N
    ],

    # channeling: sd[i][mc[i][j]] == start of operation
    [sd[i][mc[i][j]] == s[i][j] for j in M for i in N],

    # no overlap on each machine across all jobs
    [
        NoOverlap(
            origins=[sd[i][k] for i in N],
            lengths=durations[:, k],
        )
        for k in M
    ],

    # redundant: minimum completion time
    [s[i][-1] + d[i][-1] >= sum(durations[i]) for i in N],

    # makespan is the maximum end time
    [makespan >= s[i][-1] + d[i][-1] for i in N],
)

minimize(makespan)

# --- Solution output ---
if __name__ == "__main__":
    if solve() in (SAT, OPTIMUM):
        print(f"Makespan: {value(makespan)}")
        for i in N:
            print(f"Job {i}:")
            for j in M:
                start = value(s[i][j])
                machine = value(mc[i][j])
                dur = value(d[i][j])
                print(f"  Op {j} on M{machine} (dur={dur}): [{start}, {start + dur})")



"""
Sequence accessor expressions for scheduling models.

These functions return expressions that access properties of neighboring
intervals in a sequence relative to a given interval.
"""

from __future__ import annotations

from typing import TYPE_CHECKING

from pycsp3_scheduling.expressions.interval_expr import ExprType, IntervalExpr

if TYPE_CHECKING:
    from pycsp3_scheduling.variables.interval import IntervalVar
    from pycsp3_scheduling.variables.sequence import SequenceVar


def _validate_sequence_and_interval(sequence, interval: IntervalVar) -> tuple[list, int]:
    """Validate inputs and return intervals list and index."""
    from pycsp3_scheduling.variables.interval import IntervalVar
    from pycsp3_scheduling.variables.sequence import SequenceVar

    if isinstance(sequence, SequenceVar):
        intervals = sequence.intervals
    elif isinstance(sequence, (list, tuple)):
        intervals = list(sequence)
    else:
        raise TypeError(
            f"sequence must be a SequenceVar or list, got {type(sequence).__name__}"
        )

    if not isinstance(interval, IntervalVar):
        raise TypeError(
            f"interval must be an IntervalVar, got {type(interval).__name__}"
        )

    try:
        idx = intervals.index(interval)
    except ValueError:
        raise ValueError(f"interval '{interval.name}' is not in the sequence")

    return intervals, idx


# =============================================================================
# Next Interval Accessors
# =============================================================================


def start_of_next(
    sequence,
    interval: IntervalVar,
    last_value: int = 0,
    absent_value: int = 0,
) -> IntervalExpr:
    """
    Return the start time of the next interval in the sequence.

    If the given interval is last in the sequence, returns last_value.
    If the given interval is absent, returns absent_value.

    Args:
        sequence: SequenceVar or list of IntervalVar.
        interval: The reference interval.
        last_value: Value when interval is last (default: 0).
        absent_value: Value when interval is absent (default: 0).

    Returns:
        An expression representing the start of the next interval.

    Example:
        >>> seq = SequenceVar(intervals=[t1, t2, t3], name="machine")
        >>> expr = start_of_next(seq, t1)  # Returns start of t2 (or next in order)
    """
    intervals, idx = _validate_sequence_and_interval(sequence, interval)
    return IntervalExpr(
        expr_type=ExprType.START_OF,  # Placeholder - actual logic in evaluation
        interval=interval,
        absent_value=absent_value,
        value=last_value,  # Store last_value for evaluation
    )


def end_of_next(
    sequence,
    interval: IntervalVar,
    last_value: int = 0,
    absent_value: int = 0,
) -> IntervalExpr:
    """
    Return the end time of the next interval in the sequence.

    If the given interval is last in the sequence, returns last_value.
    If the given interval is absent, returns absent_value.

    Args:
        sequence: SequenceVar or list of IntervalVar.
        interval: The reference interval.
        last_value: Value when interval is last (default: 0).
        absent_value: Value when interval is absent (default: 0).

    Returns:
        An expression representing the end of the next interval.

    Example:
        >>> seq = SequenceVar(intervals=[t1, t2, t3], name="machine")
        >>> expr = end_of_next(seq, t1)  # Returns end of next interval after t1
    """
    intervals, idx = _validate_sequence_and_interval(sequence, interval)
    return IntervalExpr(
        expr_type=ExprType.END_OF,
        interval=interval,
        absent_value=absent_value,
        value=last_value,
    )


def size_of_next(
    sequence,
    interval: IntervalVar,
    last_value: int = 0,
    absent_value: int = 0,
) -> IntervalExpr:
    """
    Return the size (duration) of the next interval in the sequence.

    If the given interval is last in the sequence, returns last_value.
    If the given interval is absent, returns absent_value.

    Args:
        sequence: SequenceVar or list of IntervalVar.
        interval: The reference interval.
        last_value: Value when interval is last (default: 0).
        absent_value: Value when interval is absent (default: 0).

    Returns:
        An expression representing the size of the next interval.
    """
    intervals, idx = _validate_sequence_and_interval(sequence, interval)
    return IntervalExpr(
        expr_type=ExprType.SIZE_OF,
        interval=interval,
        absent_value=absent_value,
        value=last_value,
    )


def length_of_next(
    sequence,
    interval: IntervalVar,
    last_value: int = 0,
    absent_value: int = 0,
) -> IntervalExpr:
    """
    Return the length of the next interval in the sequence.

    If the given interval is last in the sequence, returns last_value.
    If the given interval is absent, returns absent_value.

    Args:
        sequence: SequenceVar or list of IntervalVar.
        interval: The reference interval.
        last_value: Value when interval is last (default: 0).
        absent_value: Value when interval is absent (default: 0).

    Returns:
        An expression representing the length of the next interval.
    """
    intervals, idx = _validate_sequence_and_interval(sequence, interval)
    return IntervalExpr(
        expr_type=ExprType.LENGTH_OF,
        interval=interval,
        absent_value=absent_value,
        value=last_value,
    )


def type_of_next(
    sequence,
    interval: IntervalVar,
    last_value: int = 0,
    absent_value: int = 0,
) -> IntervalExpr:
    """
    Return the type of the next interval in the sequence.

    Requires a SequenceVar with types defined.

    If the given interval is last in the sequence, returns last_value.
    If the given interval is absent, returns absent_value.

    Args:
        sequence: SequenceVar with types defined.
        interval: The reference interval.
        last_value: Value when interval is last (default: 0).
        absent_value: Value when interval is absent (default: 0).

    Returns:
        An expression representing the type of the next interval.

    Raises:
        TypeError: If sequence is not a SequenceVar or has no types.
    """
    from pycsp3_scheduling.variables.sequence import SequenceVar

    if not isinstance(sequence, SequenceVar):
        raise TypeError("type_of_next requires a SequenceVar")
    if not sequence.has_types:
        raise ValueError("type_of_next requires sequence with types defined")

    intervals, idx = _validate_sequence_and_interval(sequence, interval)
    # Create a special expression for type access
    return IntervalExpr(
        expr_type=ExprType.START_OF,  # Placeholder type
        interval=interval,
        absent_value=absent_value,
        value=last_value,
    )


# =============================================================================
# Previous Interval Accessors
# =============================================================================


def start_of_prev(
    sequence,
    interval: IntervalVar,
    first_value: int = 0,
    absent_value: int = 0,
) -> IntervalExpr:
    """
    Return the start time of the previous interval in the sequence.

    If the given interval is first in the sequence, returns first_value.
    If the given interval is absent, returns absent_value.

    Args:
        sequence: SequenceVar or list of IntervalVar.
        interval: The reference interval.
        first_value: Value when interval is first (default: 0).
        absent_value: Value when interval is absent (default: 0).

    Returns:
        An expression representing the start of the previous interval.

    Example:
        >>> seq = SequenceVar(intervals=[t1, t2, t3], name="machine")
        >>> expr = start_of_prev(seq, t2)  # Returns start of t1 (or prev in order)
    """
    intervals, idx = _validate_sequence_and_interval(sequence, interval)
    return IntervalExpr(
        expr_type=ExprType.START_OF,
        interval=interval,
        absent_value=absent_value,
        value=first_value,
    )


def end_of_prev(
    sequence,
    interval: IntervalVar,
    first_value: int = 0,
    absent_value: int = 0,
) -> IntervalExpr:
    """
    Return the end time of the previous interval in the sequence.

    If the given interval is first in the sequence, returns first_value.
    If the given interval is absent, returns absent_value.

    Args:
        sequence: SequenceVar or list of IntervalVar.
        interval: The reference interval.
        first_value: Value when interval is first (default: 0).
        absent_value: Value when interval is absent (default: 0).

    Returns:
        An expression representing the end of the previous interval.
    """
    intervals, idx = _validate_sequence_and_interval(sequence, interval)
    return IntervalExpr(
        expr_type=ExprType.END_OF,
        interval=interval,
        absent_value=absent_value,
        value=first_value,
    )


def size_of_prev(
    sequence,
    interval: IntervalVar,
    first_value: int = 0,
    absent_value: int = 0,
) -> IntervalExpr:
    """
    Return the size (duration) of the previous interval in the sequence.

    If the given interval is first in the sequence, returns first_value.
    If the given interval is absent, returns absent_value.

    Args:
        sequence: SequenceVar or list of IntervalVar.
        interval: The reference interval.
        first_value: Value when interval is first (default: 0).
        absent_value: Value when interval is absent (default: 0).

    Returns:
        An expression representing the size of the previous interval.
    """
    intervals, idx = _validate_sequence_and_interval(sequence, interval)
    return IntervalExpr(
        expr_type=ExprType.SIZE_OF,
        interval=interval,
        absent_value=absent_value,
        value=first_value,
    )


def length_of_prev(
    sequence,
    interval: IntervalVar,
    first_value: int = 0,
    absent_value: int = 0,
) -> IntervalExpr:
    """
    Return the length of the previous interval in the sequence.

    If the given interval is first in the sequence, returns first_value.
    If the given interval is absent, returns absent_value.

    Args:
        sequence: SequenceVar or list of IntervalVar.
        interval: The reference interval.
        first_value: Value when interval is first (default: 0).
        absent_value: Value when interval is absent (default: 0).

    Returns:
        An expression representing the length of the previous interval.
    """
    intervals, idx = _validate_sequence_and_interval(sequence, interval)
    return IntervalExpr(
        expr_type=ExprType.LENGTH_OF,
        interval=interval,
        absent_value=absent_value,
        value=first_value,
    )


def type_of_prev(
    sequence,
    interval: IntervalVar,
    first_value: int = 0,
    absent_value: int = 0,
) -> IntervalExpr:
    """
    Return the type of the previous interval in the sequence.

    Requires a SequenceVar with types defined.

    If the given interval is first in the sequence, returns first_value.
    If the given interval is absent, returns absent_value.

    Args:
        sequence: SequenceVar with types defined.
        interval: The reference interval.
        first_value: Value when interval is first (default: 0).
        absent_value: Value when interval is absent (default: 0).

    Returns:
        An expression representing the type of the previous interval.

    Raises:
        TypeError: If sequence is not a SequenceVar or has no types.
    """
    from pycsp3_scheduling.variables.sequence import SequenceVar

    if not isinstance(sequence, SequenceVar):
        raise TypeError("type_of_prev requires a SequenceVar")
    if not sequence.has_types:
        raise ValueError("type_of_prev requires sequence with types defined")

    intervals, idx = _validate_sequence_and_interval(sequence, interval)
    return IntervalExpr(
        expr_type=ExprType.START_OF,  # Placeholder type
        interval=interval,
        absent_value=absent_value,
        value=first_value,
    )



from pycsp3 import *
from pycsp3_scheduling import *

# Data
n_jobs, n_machines = 3, 3
durations = [[3, 2, 2], [2, 1, 4], [4, 3, 3]]
machines = [[0, 1, 2], [0, 2, 1], [1, 0, 2]]

# Create interval variables for each operation
ops = [[IntervalVar(size=durations[j][o], name=f"op_{j}_{o}")
        for o in range(n_machines)] for j in range(n_jobs)]

# Sequences for each machine
sequences = [SequenceVar(
    intervals=[ops[j][o] for j in range(n_jobs)
               for o in range(n_machines) if machines[j][o] == m],
    name=f"machine_{m}"
) for m in range(n_machines)]

# Precedence within jobs
satisfy(
    end_before_start(ops[j][o], ops[j][o+1])
    for j in range(n_jobs) for o in range(n_machines-1)
)

# No overlap on machines
satisfy(SeqNoOverlap(seq) for seq in sequences)

# Minimize makespan
minimize(Maximum(end_time(ops[j][-1]) for j in range(n_jobs)))

result = solve()
if result in [SAT, OPTIMUM]:
    for j in range(n_jobs):
        for o in range(n_machines):
            v = interval_value(ops[j][o])
            print(f"{ops[j][o].name}: start={v['start']}, end={v['end']}")
    for m in range(n_machines):
        print(f"Sequence on machine {m}: {sequences[m]}")
else: 
    print("Result: ", result)


"""Tests for state functions and constraints."""

from __future__ import annotations

import pytest

from pycsp3_scheduling import (
    IntervalVar,
    StateConstraint,
    StateFunction,
    TransitionMatrix,
    always_constant,
    always_equal,
    always_in,
    always_no_state,
)
from pycsp3_scheduling.functions.state_functions import (
    StateConstraintType,
    clear_state_function_registry,
    get_registered_state_functions,
)


@pytest.fixture(autouse=True)
def reset_state_registry():
    """Reset the state function registry before each test."""
    clear_state_function_registry()


class TestTransitionMatrix:
    """Tests for TransitionMatrix class."""

    def test_create_basic_matrix(self):
        """Test creating a basic transition matrix."""
        tm = TransitionMatrix([
            [0, 5, 10],
            [5, 0, 3],
            [10, 3, 0],
        ])

        assert tm.size == 3
        assert tm[0, 1] == 5
        assert tm[1, 2] == 3
        assert tm[2, 0] == 10

    def test_create_named_matrix(self):
        """Test creating a named transition matrix."""
        tm = TransitionMatrix(
            [[0, 1], [1, 0]],
            name="setup_times"
        )

        assert tm.name == "setup_times"
        assert tm.size == 2

    def test_matrix_must_be_square(self):
        """Test that non-square matrix raises error."""
        with pytest.raises(ValueError, match="must be square"):
            TransitionMatrix([
                [0, 1, 2],
                [1, 0],  # Wrong length
            ])

    def test_empty_matrix_raises(self):
        """Test that empty matrix raises error."""
        with pytest.raises(ValueError, match="cannot be empty"):
            TransitionMatrix([])

    def test_matrix_values_must_be_int(self):
        """Test that non-integer values raise error."""
        with pytest.raises(TypeError, match="must be integers"):
            TransitionMatrix([[0, 1.5], [1, 0]])  # type: ignore

    def test_setitem(self):
        """Test setting matrix values."""
        tm = TransitionMatrix([[0, 1], [1, 0]])
        tm[0, 1] = 5
        assert tm[0, 1] == 5

    def test_is_forbidden(self):
        """Test checking forbidden transitions."""
        tm = TransitionMatrix([
            [0, 5, -1],
            [5, 0, 3],
            [-1, 3, 0],
        ])

        assert tm.is_forbidden(0, 2)
        assert tm.is_forbidden(2, 0)
        assert not tm.is_forbidden(0, 1)

    def test_get_row(self):
        """Test getting transition times from a state."""
        tm = TransitionMatrix([
            [0, 5, 10],
            [5, 0, 3],
            [10, 3, 0],
        ])

        assert tm.get_row(1) == [5, 0, 3]

    def test_get_column(self):
        """Test getting transition times to a state."""
        tm = TransitionMatrix([
            [0, 5, 10],
            [5, 0, 3],
            [10, 3, 0],
        ])

        assert tm.get_column(1) == [5, 0, 3]

    def test_repr(self):
        """Test string representation."""
        tm = TransitionMatrix([[0, 1], [1, 0]], name="test")
        assert "test" in repr(tm)
        assert "2x2" in repr(tm)


class TestStateFunction:
    """Tests for StateFunction class."""

    def test_create_basic_state_function(self):
        """Test creating a basic state function."""
        sf = StateFunction(name="machine")

        assert sf.name == "machine"
        assert sf.transitions is None
        assert sf.initial_state is None

    def test_create_with_transitions(self):
        """Test creating a state function with transitions."""
        tm = TransitionMatrix([[0, 5], [5, 0]])
        sf = StateFunction(name="machine", transitions=tm)

        assert sf.transitions is tm
        assert sf.num_states == 2
        assert sf.states == {0, 1}

    def test_create_with_initial_state(self):
        """Test creating a state function with initial state."""
        sf = StateFunction(name="machine", initial_state=0)

        assert sf.initial_state == 0

    def test_create_with_explicit_states(self):
        """Test creating a state function with explicit states."""
        sf = StateFunction(name="machine", states={0, 1, 2, 3})

        assert sf.num_states == 4
        assert sf.states == {0, 1, 2, 3}

    def test_state_function_registration(self):
        """Test that state functions are registered."""
        sf = StateFunction(name="machine")

        registered = get_registered_state_functions()
        assert sf in registered

    def test_clear_registry(self):
        """Test clearing the registry."""
        StateFunction(name="machine")
        clear_state_function_registry()

        assert len(get_registered_state_functions()) == 0

    def test_repr(self):
        """Test string representation."""
        sf = StateFunction(name="machine")
        assert "machine" in repr(sf)


class TestAlwaysEqual:
    """Tests for always_equal constraint."""

    def test_always_equal_basic(self):
        """Test basic always_equal constraint."""
        sf = StateFunction(name="machine")
        task = IntervalVar(size=10, name="task")

        constraint = always_equal(sf, task, 2)

        assert isinstance(constraint, StateConstraint)
        assert constraint.constraint_type == StateConstraintType.ALWAYS_EQUAL
        assert constraint.state_func is sf
        assert constraint.interval is task
        assert constraint.value == 2

    def test_always_equal_alignment_options(self):
        """Test always_equal with alignment options."""
        sf = StateFunction(name="machine")
        task = IntervalVar(size=10, name="task")

        constraint = always_equal(
            sf, task, 1,
            is_start_aligned=False,
            is_end_aligned=False
        )

        assert constraint.is_start_aligned is False
        assert constraint.is_end_aligned is False

    def test_always_equal_invalid_state_func(self):
        """Test always_equal with invalid state function."""
        task = IntervalVar(size=10, name="task")

        with pytest.raises(TypeError, match="must be a StateFunction"):
            always_equal("not_a_state_func", task, 1)  # type: ignore

    def test_always_equal_invalid_interval(self):
        """Test always_equal with invalid interval."""
        sf = StateFunction(name="machine")

        with pytest.raises(TypeError, match="must be an IntervalVar"):
            always_equal(sf, "not_an_interval", 1)  # type: ignore

    def test_always_equal_invalid_value(self):
        """Test always_equal with invalid value type."""
        sf = StateFunction(name="machine")
        task = IntervalVar(size=10, name="task")

        with pytest.raises(TypeError, match="must be an int"):
            always_equal(sf, task, "not_an_int")  # type: ignore


class TestAlwaysIn:
    """Tests for always_in constraint with StateFunction."""

    def test_always_in_basic(self):
        """Test basic always_in constraint."""
        sf = StateFunction(name="machine")
        task = IntervalVar(size=10, name="task")

        constraint = always_in(sf, task, 1, 3)

        assert isinstance(constraint, StateConstraint)
        assert constraint.constraint_type == StateConstraintType.ALWAYS_IN
        assert constraint.min_value == 1
        assert constraint.max_value == 3

    def test_always_in_invalid_bounds(self):
        """Test always_in with invalid bounds."""
        sf = StateFunction(name="machine")
        task = IntervalVar(size=10, name="task")

        with pytest.raises(ValueError, match="cannot exceed"):
            always_in(sf, task, 5, 3)


class TestAlwaysConstant:
    """Tests for always_constant constraint."""

    def test_always_constant_basic(self):
        """Test basic always_constant constraint."""
        sf = StateFunction(name="machine")
        task = IntervalVar(size=10, name="task")

        constraint = always_constant(sf, task)

        assert isinstance(constraint, StateConstraint)
        assert constraint.constraint_type == StateConstraintType.ALWAYS_CONSTANT
        assert constraint.state_func is sf
        assert constraint.interval is task

    def test_always_constant_alignment_options(self):
        """Test always_constant with alignment options."""
        sf = StateFunction(name="machine")
        task = IntervalVar(size=10, name="task")

        constraint = always_constant(
            sf, task,
            is_start_aligned=False,
            is_end_aligned=True
        )

        assert constraint.is_start_aligned is False
        assert constraint.is_end_aligned is True

    def test_always_constant_invalid_inputs(self):
        """Test always_constant with invalid inputs."""
        sf = StateFunction(name="machine")
        task = IntervalVar(size=10, name="task")

        with pytest.raises(TypeError):
            always_constant("not_sf", task)  # type: ignore

        with pytest.raises(TypeError):
            always_constant(sf, "not_interval")  # type: ignore


class TestAlwaysNoState:
    """Tests for always_no_state constraint."""

    def test_always_no_state_basic(self):
        """Test basic always_no_state constraint."""
        sf = StateFunction(name="machine")
        maintenance = IntervalVar(size=10, name="maintenance")

        constraint = always_no_state(sf, maintenance)

        assert isinstance(constraint, StateConstraint)
        assert constraint.constraint_type == StateConstraintType.ALWAYS_NO_STATE
        assert constraint.state_func is sf
        assert constraint.interval is maintenance

    def test_always_no_state_alignment_options(self):
        """Test always_no_state with alignment options."""
        sf = StateFunction(name="machine")
        maintenance = IntervalVar(size=10, name="maintenance")

        constraint = always_no_state(
            sf, maintenance,
            is_start_aligned=True,
            is_end_aligned=False
        )

        assert constraint.is_start_aligned is True
        assert constraint.is_end_aligned is False


class TestStateScenarios:
    """Integration tests for realistic state function scenarios."""

    def test_machine_modes_scenario(self):
        """Test machine with different operating modes."""
        # Define transition times between modes
        # Mode 0: Idle, Mode 1: Processing, Mode 2: Setup
        transitions = TransitionMatrix([
            [0, 2, 5],   # From Idle: to Processing=2, to Setup=5
            [1, 0, 3],   # From Processing: to Idle=1, to Setup=3
            [3, 4, 0],   # From Setup: to Idle=3, to Processing=4
        ])

        machine = StateFunction(
            name="machine",
            transitions=transitions,
            initial_state=0
        )

        # Tasks that require specific modes
        task_a = IntervalVar(size=10, name="task_a")
        task_b = IntervalVar(size=5, name="task_b")

        # Both tasks need machine in Processing mode (1)
        c1 = always_equal(machine, task_a, 1)
        c2 = always_equal(machine, task_b, 1)

        assert c1.value == 1
        assert c2.value == 1
        assert machine.num_states == 3

    def test_room_configuration_scenario(self):
        """Test room with different configurations."""
        room = StateFunction(
            name="room",
            states={0, 1, 2}  # 0=empty, 1=lecture, 2=exam
        )

        lecture = IntervalVar(size=60, name="lecture")
        exam = IntervalVar(size=120, name="exam")
        cleanup = IntervalVar(size=15, name="cleanup")

        # Lecture needs room in lecture mode
        c1 = always_equal(room, lecture, 1)

        # Exam needs room in exam mode
        c2 = always_equal(room, exam, 2)

        # Cleanup - room has no specific mode
        c3 = always_no_state(room, cleanup)

        assert c1.value == 1
        assert c2.value == 2
        assert c3.constraint_type == StateConstraintType.ALWAYS_NO_STATE

    def test_worker_skills_scenario(self):
        """Test worker with different skill requirements."""
        worker_mode = StateFunction(name="worker")

        # Tasks requiring different skill levels (0-4)
        basic_task = IntervalVar(size=30, name="basic")
        advanced_task = IntervalVar(size=60, name="advanced")

        # Basic task needs skill level 1-2
        c1 = always_in(worker_mode, basic_task, 1, 2)

        # Advanced task needs skill level 3-4
        c2 = always_in(worker_mode, advanced_task, 3, 4)

        assert c1.min_value == 1
        assert c1.max_value == 2
        assert c2.min_value == 3
        assert c2.max_value == 4

    def test_constraint_repr(self):
        """Test string representations of constraints."""
        sf = StateFunction(name="machine")
        task = IntervalVar(size=10, name="task")

        c1 = always_equal(sf, task, 2)
        c2 = always_in(sf, task, 1, 3)
        c3 = always_constant(sf, task)
        c4 = always_no_state(sf, task)

        assert "always_equal" in repr(c1)
        assert "always_in" in repr(c2)
        assert "always_constant" in repr(c3)
        assert "always_no_state" in repr(c4)



"""
Helpers for bridging IntervalVar objects to pycsp3 variables.
"""

from __future__ import annotations

from typing import TYPE_CHECKING, Any

from pycsp3_scheduling.variables.interval import (
    INTERVAL_MAX,
    INTERVAL_MIN,
    IntervalVar,
    get_registered_intervals,
)

if TYPE_CHECKING:
    from pycsp3.classes.main.variables import Variable

_start_vars: dict[IntervalVar, Any] = {}
_length_vars: dict[IntervalVar, Any] = {}
_presence_vars: dict[IntervalVar, Any] = {}


def _var_id(prefix: str, interval: IntervalVar) -> str:
    return f"{prefix}{interval._id}"


def _default_horizon(intervals: list[IntervalVar]) -> int:
    horizon = 0
    for interval in intervals:
        if interval.end_max != INTERVAL_MAX:
            horizon = max(horizon, interval.end_max)
        elif interval.start_max != INTERVAL_MAX:
            horizon = max(horizon, interval.start_max + interval.length_max)
        else:
            horizon += interval.length_max
    return max(horizon, 0)


def _range_from_bounds(min_val: int, max_val: int, horizon: int) -> range:
    if max_val == INTERVAL_MAX:
        max_val = horizon
    if max_val < min_val:
        max_val = min_val
    return range(min_val, max_val + 1)


def _ensure_options(options) -> None:
    if hasattr(options, "checker"):
        return
    options.set_values(
        "data",
        "dataparser",
        "dataexport",
        "dataformat",
        "variant",
        "tocsp",
        "checker",
        "solver",
        "output",
        "suffix",
        "callback",
    )
    options.set_flags(
        "dataexport",
        "datasober",
        "solve",
        "display",
        "verbose",
        "lzma",
        "sober",
        "ev",
        "safe",
        "recognizeSlides",
        "keepHybrid",
        "keepSmartTransitions",
        "keepsum",
        "unchangescalar",
        "restrictTablesWrtDomains",
        "dontruncompactor",
        "dontcompactValues",
        "groupsumcoeffs",
        "usemeta",
        "dontuseauxcache",
        "dontadjustindexing",
        "dontbuildsimilarconstraints",
        "debug",
        "mini",
        "uncurse",
        "existbyelement",
        "safetables",
    )
    if options.checker is None:
        options.checker = "fast"


def _require_pycsp3():
    try:
        from pycsp3 import Var
        from pycsp3.dashboard import options
        from pycsp3.classes.main.variables import Variable
    except Exception as exc:  # pragma: no cover - depends on runtime environment
        raise ImportError(
            "pycsp3 is required to build scheduling constraints. "
            "Install a compatible version and Python runtime."
        ) from exc
    _ensure_options(options)
    return Var, Variable


def presence_var(interval: IntervalVar) -> Any:
    """Return (or create) a pycsp3 boolean variable for the interval presence."""
    if not interval.optional:
        # Non-optional intervals are always present - return constant 1
        return 1
    if interval in _presence_vars:
        return _presence_vars[interval]
    Var, _ = _require_pycsp3()
    var = Var(dom={0, 1}, id=_var_id("iv_p_", interval))
    _presence_vars[interval] = var
    return var


def start_var(interval: IntervalVar) -> Any:
    """Return (or create) a pycsp3 variable for the interval start time."""
    if interval in _start_vars:
        return _start_vars[interval]
    Var, _ = _require_pycsp3()
    intervals = get_registered_intervals() or [interval]
    horizon = _default_horizon(intervals)
    dom = _range_from_bounds(interval.start_min, interval.start_max, horizon)
    var = Var(dom=dom, id=_var_id("iv_s_", interval))
    _start_vars[interval] = var
    return var


def length_value(interval: IntervalVar) -> Any:
    """Return a fixed length or a pycsp3 variable for interval length."""
    if interval.length_min == interval.length_max:
        return interval.length_min
    if interval in _length_vars:
        return _length_vars[interval]
    Var, _ = _require_pycsp3()
    dom = range(interval.length_min, interval.length_max + 1)
    var = Var(dom=dom, id=_var_id("iv_l_", interval))
    _length_vars[interval] = var
    return var


def clear_pycsp3_cache() -> None:
    """Clear cached pycsp3 variables for intervals."""
    _start_vars.clear()
    _length_vars.clear()
    _presence_vars.clear()



"""
Interop helpers to build pycsp3 expressions from IntervalVar.
"""

from __future__ import annotations

from pycsp3_scheduling.constraints._pycsp3 import length_value, presence_var, start_var
from pycsp3_scheduling.variables.interval import IntervalVar


def start_time(interval: IntervalVar):
    """Return a pycsp3 variable representing the start time."""
    if not isinstance(interval, IntervalVar):
        raise TypeError("start_time expects an IntervalVar")
    return start_var(interval)


def end_time(interval: IntervalVar):
    """Return a pycsp3 expression representing the end time (start + length)."""
    if not isinstance(interval, IntervalVar):
        raise TypeError("end_time expects an IntervalVar")
    return start_var(interval) + length_value(interval)


def presence_time(interval: IntervalVar):
    """Return a pycsp3 variable representing presence (0/1) for optional intervals."""
    if not isinstance(interval, IntervalVar):
        raise TypeError("presence_time expects an IntervalVar")
    return presence_var(interval)


def interval_value(interval: IntervalVar) -> dict | None:
    """
    Extract the solution values for an interval after solving.
    
    Returns a dict with 'start', 'end', 'length', 'present' keys,
    or None if the interval is absent (for optional intervals).
    
    Args:
        interval: The interval variable to extract values from.
        
    Returns:
        Dict with start/end/length/present values, or None if absent.
        
    Example:
        >>> task = IntervalVar(size=10, name="task")
        >>> # ... add constraints and solve ...
        >>> vals = interval_value(task)
        >>> print(f"start={vals['start']}, end={vals['end']}")
    """
    from pycsp3 import value
    
    if not isinstance(interval, IntervalVar):
        raise TypeError("interval_value expects an IntervalVar")
    
    # Check presence for optional intervals
    if interval.optional:
        pres = presence_var(interval)
        if value(pres) == 0:
            return None
    
    start = value(start_var(interval))
    length_val = length_value(interval)
    if isinstance(length_val, int):
        length = length_val
    else:
        length = value(length_val)
    
    return {
        'start': start,
        'end': start + length,
        'length': length,
        'present': True,
    }



"""
Employee Scheduling Problem

A simplified employee scheduling problem where:
- Multiple employees with different skills
- Tasks require specific skills and have time windows
- Each employee can only work on one task at a time
- Some tasks require specific employees (optional intervals + alternative)
"""
from pycsp3 import *
from pycsp3_scheduling import *

# Data
# Employees with their skills (skill -> list of employee indices)
employees = ["Alice", "Bob", "Carol"]
skills = {
    "coding": [0, 1],      # Alice and Bob can code
    "testing": [1, 2],     # Bob and Carol can test
    "documentation": [0, 2],  # Alice and Carol can document
}

# Tasks: (name, duration, required_skill, earliest_start, deadline)
tasks_data = [
    ("code_module_A", 4, "coding", 0, 10),
    ("code_module_B", 3, "coding", 0, 12),
    ("test_module_A", 2, "testing", 4, 15),  # Must start after some coding done
    ("test_module_B", 2, "testing", 3, 15),
    ("write_docs", 3, "documentation", 0, 15),
]

n_tasks = len(tasks_data)
n_employees = len(employees)

# Create main interval variables for tasks
tasks = [IntervalVar(
    start=(tasks_data[t][3], tasks_data[t][4] - tasks_data[t][1]),
    size=tasks_data[t][1],
    name=tasks_data[t][0]
) for t in range(n_tasks)]

# Create alternative variables for each task-employee combination
# Only employees with the required skill can do the task
task_assignments = []
for t in range(n_tasks):
    task_name, duration, skill, earliest, deadline = tasks_data[t]
    eligible = skills[skill]
    alts = [
        IntervalVar(
            start=(earliest, deadline - duration),
            size=duration,
            optional=True,
            name=f"{task_name}_by_{employees[e]}"
        )
        for e in eligible
    ]
    task_assignments.append((eligible, alts))

# Each task must be assigned to exactly one eligible employee
satisfy(
    alternative(tasks[t], task_assignments[t][1])
    for t in range(n_tasks)
)

# Each employee can only work on one task at a time
for e in range(n_employees):
    employee_tasks = []
    for t in range(n_tasks):
        eligible, alts = task_assignments[t]
        if e in eligible:
            idx = eligible.index(e)
            employee_tasks.append(alts[idx])
    if len(employee_tasks) > 1:
        seq = SequenceVar(intervals=employee_tasks, name=f"employee_{employees[e]}")
        satisfy(SeqNoOverlap(seq))

# Precedence: testing must start after corresponding coding is done
satisfy(end_before_start(tasks[0], tasks[2]))  # test_module_A after code_module_A
satisfy(end_before_start(tasks[1], tasks[3]))  # test_module_B after code_module_B

# Minimize makespan
minimize(Maximum(end_time(t) for t in tasks))

# Solve
if solve() in [SAT, OPTIMUM]:
    print("Solution found!")
    print()
    
    # Print assignments
    for t in range(n_tasks):
        task_name = tasks_data[t][0]
        eligible, alts = task_assignments[t]
        for i, e in enumerate(eligible):
            v = interval_value(alts[i])
            if v is not None:  # This employee was assigned
                print(f"{task_name}: {employees[e]} [{v['start']}-{v['end']}]")
                break
    
    print()
    
    # Print schedule by employee
    for e in range(n_employees):
        print(f"{employees[e]}'s schedule: ", end="")
        schedule = []
        for t in range(n_tasks):
            eligible, alts = task_assignments[t]
            if e in eligible:
                idx = eligible.index(e)
                v = interval_value(alts[idx])
                if v is not None:
                    schedule.append((v['start'], tasks_data[t][0], v['end']))
        schedule.sort()
        for s, name, end in schedule:
            print(f"{name}[{s}-{end}] ", end="")
        if not schedule:
            print("(no tasks)", end="")
        print()
    
    print()
    # Calculate makespan
    makespan = max(interval_value(t)['end'] for t in tasks)
    print(f"Makespan: {makespan}")
else:
    print("No solution found")



"""Tests for IntervalVar class."""

import pytest

from pycsp3_scheduling.variables import (
    IntervalVar,
    IntervalVarArray,
    IntervalVarDict,
    clear_interval_registry,
)


@pytest.fixture(autouse=True)
def reset_registry():
    """Reset the interval registry before each test."""
    clear_interval_registry()
    yield
    clear_interval_registry()


class TestIntervalVar:
    """Tests for IntervalVar class."""

    def test_basic_creation(self):
        """Test basic interval variable creation."""
        task = IntervalVar(size=10, name="task1")
        assert task.name == "task1"
        assert task.size_min == 10
        assert task.size_max == 10
        assert task.is_fixed_size
        assert not task.optional
        assert task.is_present

    def test_size_as_range(self):
        """Test interval with size range."""
        task = IntervalVar(size=(5, 20), name="task")
        assert task.size_min == 5
        assert task.size_max == 20
        assert not task.is_fixed_size

    def test_start_end_bounds(self):
        """Test start and end bounds."""
        task = IntervalVar(
            start=(0, 100),
            end=(10, 200),
            size=10,
            name="task"
        )
        assert task.start_min == 0
        assert task.start_max == 100
        assert task.end_min == 10
        assert task.end_max == 200

    def test_optional_interval(self):
        """Test optional interval creation."""
        task = IntervalVar(size=10, optional=True, name="opt")
        assert task.optional
        assert not task.is_present

    def test_is_present_property(self):
        """Test is_present property."""
        task = IntervalVar(size=10, name="task")
        assert task.is_present
        assert not task.is_optional

        opt_task = IntervalVar(size=10, optional=True, name="opt")
        assert not opt_task.is_present
        assert opt_task.is_optional

    def test_auto_name_generation(self):
        """Test automatic name generation."""
        task1 = IntervalVar(size=10)
        task2 = IntervalVar(size=10)
        assert task1.name != task2.name
        assert task1.name.startswith("_interval_")
        assert task2.name.startswith("_interval_")

    def test_unique_ids(self):
        """Test that each interval gets a unique ID."""
        task1 = IntervalVar(size=10, name="t1")
        task2 = IntervalVar(size=10, name="t2")
        assert task1._id != task2._id

    def test_equality_by_id(self):
        """Test that equality is based on ID."""
        task1 = IntervalVar(size=10, name="task")
        task2 = IntervalVar(size=10, name="task")  # Same name, different ID
        assert task1 != task2
        assert task1 == task1

    def test_hashable(self):
        """Test that intervals are hashable."""
        task1 = IntervalVar(size=10, name="t1")
        task2 = IntervalVar(size=10, name="t2")
        s = {task1, task2}
        assert len(s) == 2
        assert task1 in s
        assert task2 in s

    def test_length_defaults_to_size(self):
        """Test that length defaults to size."""
        task = IntervalVar(size=(5, 10), name="task")
        assert task.length_min == 5
        assert task.length_max == 10

    def test_explicit_length(self):
        """Test explicit length different from size."""
        task = IntervalVar(size=10, length=(8, 12), name="task")
        assert task.size_min == 10
        assert task.size_max == 10
        assert task.length_min == 8
        assert task.length_max == 12

    def test_invalid_bounds_min_greater_than_max(self):
        """Test error on invalid bounds."""
        with pytest.raises(ValueError, match="Invalid start bounds"):
            IntervalVar(start=(100, 0), size=10, name="task")

    def test_invalid_negative_size(self):
        """Test error on negative size."""
        with pytest.raises(ValueError, match="Size cannot be negative"):
            IntervalVar(size=(-5, 10), name="task")

    def test_infeasible_bounds(self):
        """Test error on infeasible bounds."""
        with pytest.raises(ValueError, match="Infeasible bounds"):
            IntervalVar(start=(100, 100), end=(50, 50), size=10, name="task")

    def test_repr(self):
        """Test string representation."""
        task = IntervalVar(size=10, name="task1")
        repr_str = repr(task)
        assert "IntervalVar" in repr_str
        assert "task1" in repr_str
        assert "size=10" in repr_str


class TestIntervalVarArray:
    """Tests for IntervalVarArray factory function."""

    def test_1d_array(self):
        """Test 1D array creation."""
        tasks = IntervalVarArray(5, size_range=10, name="task")
        assert len(tasks) == 5
        for i, task in enumerate(tasks):
            assert isinstance(task, IntervalVar)
            assert task.name == f"task[{i}]"
            assert task.size_min == 10

    def test_2d_array(self):
        """Test 2D array creation."""
        tasks = IntervalVarArray((3, 4), size_range=(5, 20), name="op")
        assert len(tasks) == 3
        assert len(tasks[0]) == 4
        for i in range(3):
            for j in range(4):
                assert isinstance(tasks[i][j], IntervalVar)
                assert tasks[i][j].name == f"op[{i}][{j}]"
                assert tasks[i][j].size_min == 5
                assert tasks[i][j].size_max == 20

    def test_optional_array(self):
        """Test array with optional intervals."""
        tasks = IntervalVarArray(3, size_range=10, optional=True, name="task")
        for task in tasks:
            assert task.optional

    def test_with_start_end_bounds(self):
        """Test array with start/end bounds."""
        tasks = IntervalVarArray(
            3,
            start=(0, 100),
            end=(10, 200),
            size_range=10,
            name="task"
        )
        for task in tasks:
            assert task.start_min == 0
            assert task.start_max == 100
            assert task.end_min == 10
            assert task.end_max == 200


class TestIntervalVarDict:
    """Tests for IntervalVarDict factory function."""

    def test_string_keys(self):
        """Test dictionary with string keys."""
        tasks = IntervalVarDict(["A", "B", "C"], size_range=10, name="task")
        assert len(tasks) == 3
        assert "A" in tasks
        assert "B" in tasks
        assert "C" in tasks
        assert tasks["A"].name == "task[A]"
        assert tasks["A"].size_min == 10

    def test_int_keys(self):
        """Test dictionary with integer keys."""
        tasks = IntervalVarDict([1, 2, 3], size_range=15, name="job")
        assert len(tasks) == 3
        assert 1 in tasks
        assert tasks[1].name == "job[1]"

    def test_tuple_keys(self):
        """Test dictionary with tuple keys."""
        keys = [(0, 0), (0, 1), (1, 0)]
        tasks = IntervalVarDict(keys, size_range=10, name="cell")
        assert len(tasks) == 3
        assert (0, 0) in tasks
        assert tasks[(0, 1)].name == "cell[(0, 1)]"



"""
Interval expression functions for scheduling models.

These functions return expression objects that can be used in constraints
and objectives. They extract properties from interval variables.
"""

from __future__ import annotations

from dataclasses import dataclass, field
from enum import Enum, auto
from typing import TYPE_CHECKING, Any, Union

if TYPE_CHECKING:
    from pycsp3_scheduling.variables.interval import IntervalVar


class ExprType(Enum):
    """Types of interval expressions."""

    START_OF = auto()
    END_OF = auto()
    SIZE_OF = auto()
    LENGTH_OF = auto()
    PRESENCE_OF = auto()
    OVERLAP_LENGTH = auto()
    # Arithmetic combinations
    ADD = auto()
    SUB = auto()
    MUL = auto()
    DIV = auto()
    NEG = auto()
    ABS = auto()
    MIN = auto()
    MAX = auto()
    # Comparison (for constraints)
    EQ = auto()
    NE = auto()
    LT = auto()
    LE = auto()
    GT = auto()
    GE = auto()


@dataclass
class IntervalExpr:
    """
    Base class for interval-related expressions.

    These expressions represent values derived from interval variables
    that can be used in constraints and objectives.

    Attributes:
        expr_type: The type of expression.
        interval: The interval variable (if applicable).
        absent_value: Value to use when interval is absent.
        operands: Child expressions for compound expressions.
        value: Constant value (for literals).
    """

    expr_type: ExprType
    interval: IntervalVar | None = None
    absent_value: int = 0
    operands: list[IntervalExpr] = field(default_factory=list)
    value: int | None = None
    _id: int = field(default=-1, repr=False)

    def __post_init__(self) -> None:
        """Assign unique ID."""
        if self._id == -1:
            self._id = IntervalExpr._get_next_id()

    @staticmethod
    def _get_next_id() -> int:
        """Get next unique ID."""
        current = getattr(IntervalExpr, "_id_counter", 0)
        IntervalExpr._id_counter = current + 1
        return current

    # Arithmetic operators
    def __add__(self, other: Union[IntervalExpr, int]) -> IntervalExpr:
        """Add two expressions or expression and constant."""
        other_expr = _to_expr(other)
        return IntervalExpr(
            expr_type=ExprType.ADD,
            operands=[self, other_expr],
        )

    def __radd__(self, other: Union[IntervalExpr, int]) -> IntervalExpr:
        """Right addition."""
        return self.__add__(other)

    def __sub__(self, other: Union[IntervalExpr, int]) -> IntervalExpr:
        """Subtract two expressions or expression and constant."""
        other_expr = _to_expr(other)
        return IntervalExpr(
            expr_type=ExprType.SUB,
            operands=[self, other_expr],
        )

    def __rsub__(self, other: Union[IntervalExpr, int]) -> IntervalExpr:
        """Right subtraction."""
        other_expr = _to_expr(other)
        return IntervalExpr(
            expr_type=ExprType.SUB,
            operands=[other_expr, self],
        )

    def __mul__(self, other: Union[IntervalExpr, int]) -> IntervalExpr:
        """Multiply two expressions or expression and constant."""
        other_expr = _to_expr(other)
        return IntervalExpr(
            expr_type=ExprType.MUL,
            operands=[self, other_expr],
        )

    def __rmul__(self, other: Union[IntervalExpr, int]) -> IntervalExpr:
        """Right multiplication."""
        return self.__mul__(other)

    def __truediv__(self, other: Union[IntervalExpr, int]) -> IntervalExpr:
        """Divide two expressions or expression and constant."""
        other_expr = _to_expr(other)
        return IntervalExpr(
            expr_type=ExprType.DIV,
            operands=[self, other_expr],
        )

    def __neg__(self) -> IntervalExpr:
        """Negate expression."""
        return IntervalExpr(
            expr_type=ExprType.NEG,
            operands=[self],
        )

    def __abs__(self) -> IntervalExpr:
        """Absolute value of expression."""
        return IntervalExpr(
            expr_type=ExprType.ABS,
            operands=[self],
        )

    # Comparison operators (return constraint expressions)
    def __eq__(self, other: object) -> IntervalExpr:  # type: ignore[override]
        """Equality comparison."""
        if isinstance(other, (IntervalExpr, int)):
            other_expr = _to_expr(other)
            return IntervalExpr(
                expr_type=ExprType.EQ,
                operands=[self, other_expr],
            )
        return NotImplemented

    def __ne__(self, other: object) -> IntervalExpr:  # type: ignore[override]
        """Inequality comparison."""
        if isinstance(other, (IntervalExpr, int)):
            other_expr = _to_expr(other)
            return IntervalExpr(
                expr_type=ExprType.NE,
                operands=[self, other_expr],
            )
        return NotImplemented

    def __lt__(self, other: Union[IntervalExpr, int]) -> IntervalExpr:
        """Less than comparison."""
        other_expr = _to_expr(other)
        return IntervalExpr(
            expr_type=ExprType.LT,
            operands=[self, other_expr],
        )

    def __le__(self, other: Union[IntervalExpr, int]) -> IntervalExpr:
        """Less than or equal comparison."""
        other_expr = _to_expr(other)
        return IntervalExpr(
            expr_type=ExprType.LE,
            operands=[self, other_expr],
        )

    def __gt__(self, other: Union[IntervalExpr, int]) -> IntervalExpr:
        """Greater than comparison."""
        other_expr = _to_expr(other)
        return IntervalExpr(
            expr_type=ExprType.GT,
            operands=[self, other_expr],
        )

    def __ge__(self, other: Union[IntervalExpr, int]) -> IntervalExpr:
        """Greater than or equal comparison."""
        other_expr = _to_expr(other)
        return IntervalExpr(
            expr_type=ExprType.GE,
            operands=[self, other_expr],
        )

    def __hash__(self) -> int:
        """Hash based on unique ID."""
        return hash(self._id)

    def __repr__(self) -> str:
        """String representation."""
        # Check for constant value first
        if self.value is not None:
            return str(self.value)
        if self.expr_type == ExprType.START_OF:
            return f"start_of({self.interval.name if self.interval else '?'})"
        elif self.expr_type == ExprType.END_OF:
            return f"end_of({self.interval.name if self.interval else '?'})"
        elif self.expr_type == ExprType.SIZE_OF:
            return f"size_of({self.interval.name if self.interval else '?'})"
        elif self.expr_type == ExprType.LENGTH_OF:
            return f"length_of({self.interval.name if self.interval else '?'})"
        elif self.expr_type == ExprType.PRESENCE_OF:
            return f"presence_of({self.interval.name if self.interval else '?'})"
        elif self.expr_type == ExprType.OVERLAP_LENGTH:
            names = [op.interval.name if op.interval else '?' for op in self.operands]
            return f"overlap_length({names[0]}, {names[1]})"
        elif self.expr_type == ExprType.ADD:
            return f"({self.operands[0]} + {self.operands[1]})"
        elif self.expr_type == ExprType.SUB:
            return f"({self.operands[0]} - {self.operands[1]})"
        elif self.expr_type == ExprType.MUL:
            return f"({self.operands[0]} * {self.operands[1]})"
        elif self.expr_type == ExprType.DIV:
            return f"({self.operands[0]} / {self.operands[1]})"
        elif self.expr_type == ExprType.NEG:
            return f"(-{self.operands[0]})"
        elif self.expr_type == ExprType.MIN:
            return f"min({', '.join(str(op) for op in self.operands)})"
        elif self.expr_type == ExprType.MAX:
            return f"max({', '.join(str(op) for op in self.operands)})"
        elif self.expr_type == ExprType.ABS:
            return f"abs({self.operands[0]})"
        elif self.expr_type == ExprType.EQ:
            return f"({self.operands[0]} == {self.operands[1]})"
        elif self.expr_type == ExprType.NE:
            return f"({self.operands[0]} != {self.operands[1]})"
        elif self.expr_type == ExprType.LT:
            return f"({self.operands[0]} < {self.operands[1]})"
        elif self.expr_type == ExprType.LE:
            return f"({self.operands[0]} <= {self.operands[1]})"
        elif self.expr_type == ExprType.GT:
            return f"({self.operands[0]} > {self.operands[1]})"
        elif self.expr_type == ExprType.GE:
            return f"({self.operands[0]} >= {self.operands[1]})"
        return f"IntervalExpr({self.expr_type})"

    def get_intervals(self) -> list[IntervalVar]:
        """Get all interval variables referenced by this expression."""
        intervals = []
        if self.interval is not None:
            intervals.append(self.interval)
        for operand in self.operands:
            intervals.extend(operand.get_intervals())
        return intervals

    def is_comparison(self) -> bool:
        """Check if this is a comparison expression (constraint)."""
        return self.expr_type in (
            ExprType.EQ,
            ExprType.NE,
            ExprType.LT,
            ExprType.LE,
            ExprType.GT,
            ExprType.GE,
        )


def _to_expr(value: Union[IntervalExpr, int]) -> IntervalExpr:
    """Convert value to IntervalExpr."""
    if isinstance(value, IntervalExpr):
        return value
    if isinstance(value, int):
        # Create a constant expression
        return IntervalExpr(
            expr_type=ExprType.ADD,  # Dummy type for constants
            value=value,
        )
    raise TypeError(f"Cannot convert {type(value)} to IntervalExpr")


# ============================================================================
# Public API Functions
# ============================================================================


def start_of(interval: IntervalVar, absent_value: int = 0) -> IntervalExpr:
    """
    Return an expression representing the start time of an interval.

    If the interval is absent (optional and not selected), returns absent_value.

    Args:
        interval: The interval variable.
        absent_value: Value to return if interval is absent (default: 0).

    Returns:
        An expression representing the start time.

    Example:
        >>> task = IntervalVar(size=10, name="task")
        >>> expr = start_of(task)
        >>> # Can be used in constraints: start_of(task) >= 5
    """
    return IntervalExpr(
        expr_type=ExprType.START_OF,
        interval=interval,
        absent_value=absent_value,
    )


def end_of(interval: IntervalVar, absent_value: int = 0) -> IntervalExpr:
    """
    Return an expression representing the end time of an interval.

    If the interval is absent (optional and not selected), returns absent_value.

    FIXME: end_of() still returns an internal IntervalExpr; for pycsp3 objectives use end_time() for now.

    Args:
        interval: The interval variable.
        absent_value: Value to return if interval is absent (default: 0).

    Returns:
        An expression representing the end time.

    Example:
        >>> task = IntervalVar(size=10, name="task")
        >>> expr = end_of(task)
        >>> # Can be used in constraints: end_of(task) <= 100
    """
    return IntervalExpr(
        expr_type=ExprType.END_OF,
        interval=interval,
        absent_value=absent_value,
    )


def size_of(interval: IntervalVar, absent_value: int = 0) -> IntervalExpr:
    """
    Return an expression representing the size (duration) of an interval.

    If the interval is absent (optional and not selected), returns absent_value.

    Args:
        interval: The interval variable.
        absent_value: Value to return if interval is absent (default: 0).

    Returns:
        An expression representing the size.

    Example:
        >>> task = IntervalVar(size=(5, 20), name="task")
        >>> expr = size_of(task)
        >>> # Can be used in constraints: size_of(task) >= 10
    """
    return IntervalExpr(
        expr_type=ExprType.SIZE_OF,
        interval=interval,
        absent_value=absent_value,
    )


def length_of(interval: IntervalVar, absent_value: int = 0) -> IntervalExpr:
    """
    Return an expression representing the length of an interval.

    Length can differ from size when intensity functions are used.
    If the interval is absent, returns absent_value.

    Args:
        interval: The interval variable.
        absent_value: Value to return if interval is absent (default: 0).

    Returns:
        An expression representing the length.

    Example:
        >>> task = IntervalVar(size=10, length=(8, 12), name="task")
        >>> expr = length_of(task)
    """
    return IntervalExpr(
        expr_type=ExprType.LENGTH_OF,
        interval=interval,
        absent_value=absent_value,
    )


def presence_of(interval: IntervalVar) -> IntervalExpr:
    """
    Return a boolean expression representing whether the interval is present.

    For mandatory intervals, this is always true.
    For optional intervals, this is a decision variable.

    Args:
        interval: The interval variable.

    Returns:
        A boolean expression (0 or 1) for presence.

    Example:
        >>> task = IntervalVar(size=10, optional=True, name="task")
        >>> expr = presence_of(task)
        >>> # Can be used: presence_of(task) == 1 means task is selected
    """
    return IntervalExpr(
        expr_type=ExprType.PRESENCE_OF,
        interval=interval,
        absent_value=0,  # Not applicable for presence
    )


def overlap_length(
    interval1: IntervalVar,
    interval2: IntervalVar,
    absent_value: int = 0,
) -> IntervalExpr:
    """
    Return an expression for the overlap length between two intervals.

    The overlap is max(0, min(end1, end2) - max(start1, start2)).
    If either interval is absent, returns absent_value.

    Args:
        interval1: First interval variable.
        interval2: Second interval variable.
        absent_value: Value to return if either interval is absent.

    Returns:
        An expression representing the overlap length.

    Example:
        >>> task1 = IntervalVar(size=10, name="task1")
        >>> task2 = IntervalVar(size=15, name="task2")
        >>> expr = overlap_length(task1, task2)
        >>> # expr == 0 means no overlap
    """
    # Create placeholder expressions for the two intervals
    expr1 = IntervalExpr(
        expr_type=ExprType.START_OF,
        interval=interval1,
    )
    expr2 = IntervalExpr(
        expr_type=ExprType.START_OF,
        interval=interval2,
    )
    return IntervalExpr(
        expr_type=ExprType.OVERLAP_LENGTH,
        operands=[expr1, expr2],
        absent_value=absent_value,
    )


# ============================================================================
# Utility Functions
# ============================================================================


def expr_min(*args: Union[IntervalExpr, int]) -> IntervalExpr:
    """
    Return the minimum of multiple expressions.

    Args:
        *args: Expressions or integers to take minimum of.

    Returns:
        An expression representing the minimum.
    """
    if len(args) < 2:
        raise ValueError("expr_min requires at least 2 arguments")
    exprs = [_to_expr(a) for a in args]
    return IntervalExpr(
        expr_type=ExprType.MIN,
        operands=exprs,
    )


def expr_max(*args: Union[IntervalExpr, int]) -> IntervalExpr:
    """
    Return the maximum of multiple expressions.

    Args:
        *args: Expressions or integers to take maximum of.

    Returns:
        An expression representing the maximum.
    """
    if len(args) < 2:
        raise ValueError("expr_max requires at least 2 arguments")
    exprs = [_to_expr(a) for a in args]
    return IntervalExpr(
        expr_type=ExprType.MAX,
        operands=exprs,
    )
